"""
Copy from pegasus and cellual

"""


import os
import logging
from functools import reduce
from random import seed, sample
from typing import Union, Optional, Sequence, Tuple, List, Dict
import pandas as pd 
import anndata
import numpy as np 
import scanpy as sc
import matplotlib.pyplot as plt 
from matplotlib.backends.backend_pdf import PdfPages 
from matplotlib import rcParams
import seaborn as sns
from scipy.sparse import issparse

from .._settings import settings,print_gpu_usage_color,EMOJI,add_reference



def mads(meta, cov, nmads=5, lt=None):
    """
    Given a certain array, it calculate its Median Absolute Deviation (MAD).
    """
    x = meta[cov]
    mad = np.median(np.absolute(x - np.median(x)))
    t1 = np.median(x) - (nmads * mad)
    t1 = t1 if t1 > 0 else lt[cov]
    t2 = np.median(x) + (nmads * mad)
    return t1, t2

def mads_test(meta, cov, nmads=5, lt=None):
    """
    Given a certain array, it returns a boolean array with True values only at indeces 
    from entries within x < n_mads*mad and x > n_mads*mad.
    """
    tresholds = mads(meta, cov, nmads=nmads, lt=lt)
    return (meta[cov] > tresholds[0]) & (meta[cov] < tresholds[1])

def quantity_control(adatas, mode='seurat', min_cells=3, min_genes=200,\
    nmads=5, path_viz=None, tresh=None):
    """
    Perform quality control on a dictionary of AnnData objects.
    
    This function calculates several QC metrics, including mitochondrial percentage, nUMIs, 
    and detected genes, and produces several plots visualizing the QC metrics for each sample. 
    The function performs doublet detection using scrublet and filtering using either 
    Seurat or MADs. The function returns a merged AnnData object with cells that passed QC filters
    and a list of cells that did not pass QC on all samples.

    Parameters
    ----------
    adatas : dict
        A dictionary of AnnData objects, one for each sample.
    mode : str, optional
        The filtering method to use. Valid options are 'seurat' and 'mads'. Default is 'seurat'.
    min_cells : int, optional
        The minimum number of cells for a sample to pass QC. Default is 3.
    min_genes : int, optional
        The minimum number of genes for a cell to pass QC. Default is 200.
    nmads : int, optional
        The number of MADs to use for MADs filtering. Default is 5.
    path_viz : str, optional
        The path to save the QC plots. Default is None.
    tresh : dict, optional
        A dictionary of QC thresholds. The keys should be 'mito_perc', 'nUMIs', 
        and 'detected_genes'.
        Only used if mode is 'seurat'. Default is None.

    Returns
    -------
    adata : AnnData
        An AnnData object containing cells that passed QC filters.
    removed_cells : list
        List of cells that did not pass QC on all samples.
    """
    # Logging
    logger = logging.getLogger("my_logger") 
    if tresh is None:
        tresh={'mito_perc': 0.15, 'nUMIs': 500, 'detected_genes': 250}
    # For each adata, produce a figure
    # with PdfPages(path_viz + 'original_QC_by_sample.pdf') as pdf:
    removed_cells = []
    for s, adata in adatas.items():


        print(f'Sample {s} QC...')

        # QC metrics
        print('Calculate QC metrics')
        adata.var_names_make_unique()
        adata.var["mt"] = adata.var_names.str.startswith("MT-")
        if issparse(adata.X):
            adata.obs['nUMIs'] = adata.X.toarray().sum(axis=1)
            adata.obs['mito_perc'] = adata[:, adata.var["mt"]].X.toarray().sum(axis=1) / \
            adata.obs['nUMIs'].values
            adata.obs['detected_genes'] = (adata.X.toarray() > 0).sum(axis=1)
        else:
            adata.obs['nUMIs'] = adata.X.sum(axis=1)
            adata.obs['mito_perc'] = adata[:, adata.var["mt"]].X.sum(axis=1) / \
            adata.obs['nUMIs'].values
            adata.obs['detected_genes'] = (adata.X > 0).sum(axis=1)
        adata.obs['cell_complexity'] = adata.obs['detected_genes'] / adata.obs['nUMIs']
        print('End calculation of QC metrics.')

        # Original QC plot
        n0 = adata.shape[0]
        print(f'Original cell number: {n0}')

        # Post doublets removal QC plot
        print('Begin of post doublets removal and QC plot')
        sc.pp.scrublet(adata, random_state=1234)
        adata_remove = adata[adata.obs['predicted_doublet'], :]
        removed_cells.extend(list(adata_remove.obs_names))
        adata = adata[~adata.obs['predicted_doublet'], :]
        n1 = adata.shape[0]
        print(f'Cells retained after scrublet: {n1}, {n0-n1} removed.')
        print('End of post doublets removal and QC plots.')

        # Post seurat or mads filtering QC plot

        # Filters
        print('Filters application (seurat or mads)')
        if mode == 'seurat':
            adata.obs['passing_mt'] = adata.obs['mito_perc'] < tresh['mito_perc']
            adata.obs['passing_nUMIs'] = adata.obs['nUMIs'] > tresh['nUMIs']
            adata.obs['passing_ngenes'] = adata.obs['detected_genes'] > tresh['detected_genes']
        elif mode == 'mads':
            adata.obs['passing_mt'] = adata.obs['mito_perc'] < tresh['mito_perc']
            adata.obs['passing_nUMIs'] = mads_test(adata.obs, 'nUMIs', nmads=nmads, lt=tresh)
            adata.obs['passing_ngenes'] = mads_test(adata.obs, 'detected_genes', \
            nmads=nmads, lt=tresh)

        # Report
        if mode == 'seurat':
            print(f'Lower treshold, nUMIs: {tresh["nUMIs"]}; filtered-out-cells: \
            {n1-np.sum(adata.obs["passing_nUMIs"])}')
            print(f'Lower treshold, n genes: {tresh["detected_genes"]}; \
            filtered-out-cells: {n1-np.sum(adata.obs["passing_ngenes"])}')
            print(f'Lower treshold, mito %: {tresh["mito_perc"]}; \
            filtered-out-cells: {n1-np.sum(adata.obs["passing_mt"])}')
        elif mode == 'mads':
            nUMIs_t = mads(adata.obs, 'nUMIs', nmads=nmads, lt=tresh)
            n_genes_t = mads(adata.obs, 'detected_genes', nmads=nmads, lt=tresh)
            print(f'Tresholds used, nUMIs: ({nUMIs_t[0]}, {nUMIs_t[1]}); \
            filtered-out-cells: {n1-np.sum(adata.obs["passing_nUMIs"])}')
            print(f'Tresholds used, n genes: ({n_genes_t[0]}, {n_genes_t[1]}); \
                filtered-out-cells: {n1-np.sum(adata.obs["passing_ngenes"])}')
            print(f'Lower treshold, mito %: {tresh["mito_perc"]}; \
            filtered-out-cells: {n1-np.sum(adata.obs["passing_mt"])}')
        print('Filters applicated.')

        # QC plot
        QC_test = (adata.obs['passing_mt']) & (adata.obs['passing_nUMIs']) & \
        (adata.obs['passing_ngenes'])
        removed = QC_test.loc[lambda x : x is False]
        removed_cells.extend(list(removed.index.values))
        print(f'Total cell filtered out with this last --mode {mode} QC (and its \
        chosen options): {n1-np.sum(QC_test)}')
        adata = adata[QC_test, :]
        n2 = adata.shape[0]
        # Store cleaned adata
        print(f'Cells retained after scrublet and {mode} filtering: {n2}, {n0-n2} removed.')
        adatas[s] = adata
        print(adatas[s])


    # Concenate
    universe = sorted(
        list(reduce(lambda x,y: x&y, [ set(adatas[k].var_names) for k in adatas ]))
    )
    seed(1234)
    universe = sample(universe, len(universe))
    adata = anndata.concat([ adatas[k][:, universe] for k in adatas ], axis=0)

    # Last gene and cell filter
    sc.pp.filter_cells(adata, min_genes=min_genes)
    sc.pp.filter_genes(adata, min_cells=min_cells)

    return adata, removed_cells


def qc(adata,**kwargs):
    '''
    qc
    '''


    if settings.mode == 'gpu':
        print(f"{EMOJI['gpu']} Using RAPIDS GPU to calculate QC...")
        return qc_gpu(adata,**kwargs)
    elif settings.mode == 'cpu-gpu-mixed':
        print(f"{EMOJI['mixed']} Using torch CPU/GPU mixed mode...")
        print_gpu_usage_color()
        return qc_cpu_gpu_mixed(adata,**kwargs)
    else:
        print(f"{EMOJI['cpu']} Using torch CPU mode...")
        return qc_cpu(adata,**kwargs)
    

def qc_cpu_gpu_mixed(adata:anndata.AnnData, mode='seurat',
       min_cells=3, min_genes=200, nmads=5,
       max_cells_ratio=1,max_genes_ratio=1,
       batch_key=None,doublets=True,doublets_method='scrublet',
       path_viz=None, tresh=None,mt_startswith='MT-',mt_genes=None,use_gpu=True):
    """
    Perform quality control on a dictionary of AnnData objects.
    
    This function calculates several QC metrics, including mitochondrial percentage, nUMIs, 
    and detected genes, and produces several plots visualizing the QC metrics for each sample. 
    The function performs doublet detection using scrublet and filtering using either 
    Seurat or MADs. The function returns a merged AnnData object with cells that passed QC filters
    and a list of cells that did not pass QC on all samples.

    Arguments:
        adatas : AnnData objects
        mode : The filtering method to use. Valid options are 'seurat' 
        and 'mads'. Default is 'seurat'.
        min_cells : The minimum number of cells for a sample to pass QC. Default is 3.
        min_genes : The minimum number of genes for a cell to pass QC. Default is 200.
        max_cells_ratio : The maximum number of cells ratio for a sample to pass QC. Default is 1.
        max_genes_ratio : The maximum number of genes ratio for a cell to pass QC. Default is 1.
        nmads : The number of MADs to use for MADs filtering. Default is 5.
        path_viz : The path to save the QC plots. Default is None.
        tresh : A dictionary of QC thresholds. The keys should be 'mito_perc', 
        'nUMIs', and 'detected_genes'.
            Only used if mode is 'seurat'. Default is None.
        mt_startswith : The prefix of mitochondrial genes. Default is 'MT-'.
        mt_genes : The list of mitochondrial genes. Default is None. 
        if mt_genes is not None, mt_startswith will be ignored.

    Returns:
        adata : An AnnData object containing cells that passed QC filters.

    """
    # Logging
    if tresh is None:
        tresh={'mito_perc': 0.15, 'nUMIs': 500, 'detected_genes': 250}
    # For each adata, produce a figure
    # with PdfPages(path_viz + 'original_QC_by_sample.pdf') as pdf:
    
    removed_cells = []

    # QC metrics
    print('Calculate QC metrics')
    adata.var_names_make_unique()
    if mt_genes is not None:
        adata.var['mt']=False
        adata.var.loc[list(set(adata.var_names) & set(mt_genes)),'mt']=True
    else:
        adata.var["mt"] = adata.var_names.str.startswith(mt_startswith)
    if issparse(adata.X):
        adata.obs['nUMIs'] = np.array(adata.X.sum(axis=1)).reshape(-1)
        adata.obs['mito_perc'] = np.array(adata[:, adata.var["mt"]].X.sum(axis=1)).reshape(-1) / \
        adata.obs['nUMIs'].values
        adata.obs['detected_genes'] = adata.X.getnnz(axis=1)
    else:
        adata.obs['nUMIs'] = adata.X.sum(axis=1)
        adata.obs['mito_perc'] = adata[:, adata.var["mt"] is True].X.sum(axis=1) / \
        adata.obs['nUMIs'].values
        adata.obs['detected_genes'] = np.count_nonzero(adata.X, axis=1)
    adata.obs['cell_complexity'] = adata.obs['detected_genes'] / adata.obs['nUMIs']
    print('End calculation of QC metrics.')

    # Original QC plot
    n0 = adata.shape[0]
    print(f'Original cell number: {n0}')
    n1 = n0

    if doublets is True:
        if doublets_method=='scrublet':
            # Post doublets removal QC plot
            print('!!!It should be noted that the `scrublet` detection is too old and \
            may not work properly.!!!')
            print('!!!if you want to use novel doublet detection, \
            please set `doublets_method=sccomposite`!!!')
            print('Begin of post doublets removal and QC plot using`scrublet`')
            from ._scrublet import scrublet
            scrublet(adata, random_state=1234,batch_key=batch_key,use_gpu=use_gpu)

            adata_remove = adata[adata.obs['predicted_doublet'], :]
            removed_cells.extend(list(adata_remove.obs_names))
            adata = adata[~adata.obs['predicted_doublet'], :]
            n1 = adata.shape[0]
            print(f'Cells retained after scrublet: {n1}, {n0-n1} removed.')
            print('End of post doublets removal and QC plots.')
        elif doublets_method=='sccomposite':
            print('!!!It should be noted that the `sccomposite` will remove more cells than \
            `scrublet`!!!')
            print('Begin of post doublets removal and QC plot using `sccomposite`')
            adata.obs['sccomposite_doublet']=0
            adata.obs['sccomposite_consistency']=0
            if batch_key is None:
                from ._sccomposite import composite_rna
                multiplet_classification, consistency = composite_rna(adata)
                adata.obs['sccomposite_doublet']=multiplet_classification
                adata.obs['sccomposite_consistency']=consistency
            else:
                for batch in adata.obs[batch_key].unique():
                    from ._sccomposite import composite_rna
                    adata_batch=adata[adata.obs[batch_key]==batch]
                    multiplet_classification, consistency = composite_rna(adata_batch)
                    adata.obs.loc[adata_batch.obs.index,'sccomposite_doublet']=\
                    multiplet_classification
                    adata.obs.loc[adata_batch.obs.index,'sccomposite_consistency']=consistency
            adata_remove = adata[adata.obs['sccomposite_doublet']!=0, :]
            removed_cells.extend(list(adata_remove.obs_names))
            adata = adata[adata.obs['sccomposite_doublet']==0, :]
            n1 = adata.shape[0]
            print(f'Cells retained after sccomposite: {n1}, {n0-n1} removed.')
            print('End of post sccomposite removal and QC plots.')
    else:
        n1 = adata.shape[0]

    # Post seurat or mads filtering QC plot

    # Filters
    print('Filters application (seurat or mads)')
    if mode == 'seurat':
        adata.obs['passing_mt'] = adata.obs['mito_perc'] < tresh['mito_perc']
        adata.obs['passing_nUMIs'] = adata.obs['nUMIs'] > tresh['nUMIs']
        adata.obs['passing_ngenes'] = adata.obs['detected_genes'] > tresh['detected_genes']
    elif mode == 'mads':
        adata.obs['passing_mt'] = adata.obs['mito_perc'] < tresh['mito_perc']
        adata.obs['passing_nUMIs'] = mads_test(adata.obs, 'nUMIs', nmads=nmads, lt=tresh)
        adata.obs['passing_ngenes'] = mads_test(adata.obs, 'detected_genes', nmads=nmads, lt=tresh)

    # Report
    if mode == 'seurat':
        print(f'Lower treshold, nUMIs: {tresh["nUMIs"]}; filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_nUMIs"])}')
        print(f'Lower treshold, n genes: {tresh["detected_genes"]}; filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_ngenes"])}')
        print(f'Lower treshold, mito %: {tresh["mito_perc"]}; filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_mt"])}')
    elif mode == 'mads':
        nUMIs_t = mads(adata.obs, 'nUMIs', nmads=nmads, lt=tresh)
        n_genes_t = mads(adata.obs, 'detected_genes', nmads=nmads, lt=tresh)
        print(f'Tresholds used, nUMIs: ({nUMIs_t[0]}, {nUMIs_t[1]}); filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_nUMIs"])}')
        print(f'Tresholds used, n genes: ({n_genes_t[0]}, {n_genes_t[1]}); filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_ngenes"])}')
        print(f'Lower treshold, mito %: {tresh["mito_perc"]}; filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_mt"])}')
    print('Filters applicated.')

    # QC plot
    QC_test = (adata.obs['passing_mt']) & (adata.obs['passing_nUMIs']) & (adata.obs['passing_ngenes'])
    removed = QC_test.loc[lambda x : x == False]
    removed_cells.extend(list(removed.index.values))
    print(f'Total cell filtered out with this last --mode {mode} QC (and its \
    chosen options): {n1-np.sum(QC_test)}')
    adata = adata[QC_test, :]
    n2 = adata.shape[0]
    # Store cleaned adata
    print(f'Cells retained after scrublet and {mode} filtering: {n2}, {n0-n2} removed.')

    # Last gene and cell filter
    sc.pp.filter_cells(adata, min_genes=min_genes)
    sc.pp.filter_genes(adata, min_cells=min_cells)
    sc.pp.filter_cells(adata, max_genes=max_genes_ratio*adata.shape[1])
    sc.pp.filter_genes(adata, max_cells=max_cells_ratio*adata.shape[0])

    if 'status' not in adata.uns.keys():
        adata.uns['status'] = {}
    if 'status_args' not in adata.uns.keys():
        adata.uns['status_args'] = {}

    adata.uns['status']['qc']=True
    adata.uns['status_args']['qc'] = {
        'mode': mode,
        'min_cells': min_cells,
        'min_genes': min_genes,
        'nmads': nmads,
        'max_cells_ratio': max_cells_ratio,
        'max_genes_ratio': max_genes_ratio,
        'batch_key': batch_key,
        'doublets': doublets,
        'doublets_method': doublets_method,
        'path_viz': path_viz,
        'mito_perc': tresh['mito_perc'],
        'nUMIs': tresh['nUMIs'],
        'detected_genes': tresh['detected_genes'],
    }

    return adata


def qc_cpu(adata:anndata.AnnData, mode='seurat',
       min_cells=3, min_genes=200, nmads=5,
       max_cells_ratio=1,max_genes_ratio=1,
       batch_key=None,doublets=True,doublets_method='scrublet',
       path_viz=None, tresh=None,mt_startswith='MT-',mt_genes=None):
    """
    Perform quality control on a dictionary of AnnData objects.
    
    This function calculates several QC metrics, including mitochondrial percentage, nUMIs, 
    and detected genes, and produces several plots visualizing the QC metrics for each sample. 
    The function performs doublet detection using scrublet and filtering using either 
    Seurat or MADs. The function returns a merged AnnData object with cells that passed QC filters
    and a list of cells that did not pass QC on all samples.

    Arguments:
        adatas : AnnData objects
        mode : The filtering method to use. Valid options are 'seurat' 
        and 'mads'. Default is 'seurat'.
        min_cells : The minimum number of cells for a sample to pass QC. Default is 3.
        min_genes : The minimum number of genes for a cell to pass QC. Default is 200.
        max_cells_ratio : The maximum number of cells ratio for a sample to pass QC. Default is 1.
        max_genes_ratio : The maximum number of genes ratio for a cell to pass QC. Default is 1.
        nmads : The number of MADs to use for MADs filtering. Default is 5.
        path_viz : The path to save the QC plots. Default is None.
        tresh : A dictionary of QC thresholds. The keys should be 'mito_perc', 
        'nUMIs', and 'detected_genes'.
            Only used if mode is 'seurat'. Default is None.
        mt_startswith : The prefix of mitochondrial genes. Default is 'MT-'.
        mt_genes : The list of mitochondrial genes. Default is None. 
        if mt_genes is not None, mt_startswith will be ignored.

    Returns:
        adata : An AnnData object containing cells that passed QC filters.

    """
    # Logging
    if tresh is None:
        tresh={'mito_perc': 0.15, 'nUMIs': 500, 'detected_genes': 250}
    # For each adata, produce a figure
    # with PdfPages(path_viz + 'original_QC_by_sample.pdf') as pdf:
    removed_cells = []

    # QC metrics
    print('Calculate QC metrics')
    adata.var_names_make_unique()
    if mt_genes is not None:
        adata.var['mt']=False
        adata.var.loc[list(set(adata.var_names) & set(mt_genes)),'mt']=True
    else:
        adata.var["mt"] = adata.var_names.str.startswith(mt_startswith)
    if issparse(adata.X):
        adata.obs['nUMIs'] = np.array(adata.X.sum(axis=1)).reshape(-1)
        adata.obs['mito_perc'] = np.array(adata[:, adata.var["mt"]].X.sum(axis=1)).reshape(-1) / \
        adata.obs['nUMIs'].values
        adata.obs['detected_genes'] = adata.X.getnnz(axis=1)
    else:
        adata.obs['nUMIs'] = adata.X.sum(axis=1)
        adata.obs['mito_perc'] = adata[:, adata.var["mt"] is True].X.sum(axis=1) / \
        adata.obs['nUMIs'].values
        adata.obs['detected_genes'] = np.count_nonzero(adata.X, axis=1)
    adata.obs['cell_complexity'] = adata.obs['detected_genes'] / adata.obs['nUMIs']
    print('End calculation of QC metrics.')

    # Original QC plot
    n0 = adata.shape[0]
    print(f'Original cell number: {n0}')

    if doublets is True:
        if doublets_method=='scrublet':
            # Post doublets removal QC plot
            print('!!!It should be noted that the `scrublet` detection is too old and \
            may not work properly.!!!')
            print('!!!if you want to use novel doublet detection, \
            please set `doublets_method=sccomposite`!!!')
            print('Begin of post doublets removal and QC plot using`scrublet`')
            sc.pp.scrublet(adata, random_state=1234,batch_key=batch_key)

            adata_remove = adata[adata.obs['predicted_doublet'], :]
            removed_cells.extend(list(adata_remove.obs_names))
            adata = adata[~adata.obs['predicted_doublet'], :]
            n1 = adata.shape[0]
            print(f'Cells retained after scrublet: {n1}, {n0-n1} removed.')
            print('End of post doublets removal and QC plots.')
        elif doublets_method=='sccomposite':
            print('!!!It should be noted that the `sccomposite` will remove more cells than \
            `scrublet`!!!')
            print('Begin of post doublets removal and QC plot using `sccomposite`')
            adata.obs['sccomposite_doublet']=0
            adata.obs['sccomposite_consistency']=0
            if batch_key is None:
                from ._sccomposite import composite_rna
                multiplet_classification, consistency = composite_rna(adata)
                adata.obs['sccomposite_doublet']=multiplet_classification
                adata.obs['sccomposite_consistency']=consistency
            else:
                for batch in adata.obs[batch_key].unique():
                    from ._sccomposite import composite_rna
                    adata_batch=adata[adata.obs[batch_key]==batch]
                    multiplet_classification, consistency = composite_rna(adata_batch)
                    adata.obs.loc[adata_batch.obs.index,'sccomposite_doublet']=\
                    multiplet_classification
                    adata.obs.loc[adata_batch.obs.index,'sccomposite_consistency']=consistency
            adata_remove = adata[adata.obs['sccomposite_doublet']!=0, :]
            removed_cells.extend(list(adata_remove.obs_names))
            adata = adata[adata.obs['sccomposite_doublet']==0, :]
            n1 = adata.shape[0]
            print(f'Cells retained after sccomposite: {n1}, {n0-n1} removed.')
            print('End of post sccomposite removal and QC plots.')
    else:
        n1 = adata.shape[0]
    # Fix bug where n1 is not defined when doublets=False
    if not doublets:
        n1 = n0


    # Post seurat or mads filtering QC plot

    # Filters
    print('Filters application (seurat or mads)')
    if mode == 'seurat':
        adata.obs['passing_mt'] = adata.obs['mito_perc'] < tresh['mito_perc']
        adata.obs['passing_nUMIs'] = adata.obs['nUMIs'] > tresh['nUMIs']
        adata.obs['passing_ngenes'] = adata.obs['detected_genes'] > tresh['detected_genes']
    elif mode == 'mads':
        adata.obs['passing_mt'] = adata.obs['mito_perc'] < tresh['mito_perc']
        adata.obs['passing_nUMIs'] = mads_test(adata.obs, 'nUMIs', nmads=nmads, lt=tresh)
        adata.obs['passing_ngenes'] = mads_test(adata.obs, 'detected_genes', nmads=nmads, lt=tresh)

    # Report
    if mode == 'seurat':
        print(f'Lower treshold, nUMIs: {tresh["nUMIs"]}; filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_nUMIs"])}')
        print(f'Lower treshold, n genes: {tresh["detected_genes"]}; filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_ngenes"])}')
        print(f'Lower treshold, mito %: {tresh["mito_perc"]}; filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_mt"])}')
    elif mode == 'mads':
        nUMIs_t = mads(adata.obs, 'nUMIs', nmads=nmads, lt=tresh)
        n_genes_t = mads(adata.obs, 'detected_genes', nmads=nmads, lt=tresh)
        print(f'Tresholds used, nUMIs: ({nUMIs_t[0]}, {nUMIs_t[1]}); filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_nUMIs"])}')
        print(f'Tresholds used, n genes: ({n_genes_t[0]}, {n_genes_t[1]}); filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_ngenes"])}')
        print(f'Lower treshold, mito %: {tresh["mito_perc"]}; filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_mt"])}')
    print('Filters applicated.')

    # QC plot
    QC_test = (adata.obs['passing_mt']) & (adata.obs['passing_nUMIs']) & (adata.obs['passing_ngenes'])
    removed = QC_test.loc[lambda x : x == False]
    removed_cells.extend(list(removed.index.values))
    print(f'Total cell filtered out with this last --mode {mode} QC (and its \
    chosen options): {n1-np.sum(QC_test)}')
    adata = adata[QC_test, :]
    n2 = adata.shape[0]
    # Store cleaned adata
    print(f'Cells retained after scrublet and {mode} filtering: {n2}, {n0-n2} removed.')

    # Last gene and cell filter
    sc.pp.filter_cells(adata, min_genes=min_genes)
    sc.pp.filter_genes(adata, min_cells=min_cells)
    sc.pp.filter_cells(adata, max_genes=max_genes_ratio*adata.shape[1])
    sc.pp.filter_genes(adata, max_cells=max_cells_ratio*adata.shape[0])

    if 'status' not in adata.uns.keys():
        adata.uns['status'] = {}
    if 'status_args' not in adata.uns.keys():
        adata.uns['status_args'] = {}

    adata.uns['status']['qc']=True
    adata.uns['status_args']['qc'] = {
        'mode': mode,
        'min_cells': min_cells,
        'min_genes': min_genes,
        'nmads': nmads,
        'max_cells_ratio': max_cells_ratio,
        'max_genes_ratio': max_genes_ratio,
        'batch_key': batch_key,
        'doublets': doublets,
        'doublets_method': doublets_method,
        'path_viz': path_viz,
        'mito_perc': tresh['mito_perc'],
        'nUMIs': tresh['nUMIs'],
        'detected_genes': tresh['detected_genes'],
    }
    add_reference(adata,'qc','QC with scanpy')
    return adata


def qc_gpu(adata, mode='seurat',
       min_cells=3, min_genes=200, nmads=5,
       max_cells_ratio=1,max_genes_ratio=1,
       batch_key=None,doublets=True,doublets_method='scrublet',
       path_viz=None, tresh=None,mt_startswith='MT-',mt_genes=None):
    '''
    qc
    
    '''
    import rapids_singlecell as rsc
     # Logging
    if tresh is None:
        tresh={'mito_perc': 0.15, 'nUMIs': 500, 'detected_genes': 250}
    # For each adata, produce a figure
    # with PdfPages(path_viz + 'original_QC_by_sample.pdf') as pdf:
    removed_cells = []
    rsc.get.anndata_to_GPU(adata)
    # QC metrics
    print('Calculate QC metrics')
    adata.var_names_make_unique()
    if mt_genes is not None:
        adata.var['mt']=False
        adata.var.loc[list(set(adata.var_names) & set(mt_genes)),'mt']=True
    else:
        rsc.pp.flag_gene_family(adata, gene_family_name="mt", gene_family_prefix=mt_startswith)
    rsc.pp.calculate_qc_metrics(adata, qc_vars=["mt"])
    adata.obs['nUMIs'] = adata.obs['total_counts']
    adata.obs['mito_perc'] = adata.obs['pct_counts_mt']/100
    adata.obs['detected_genes'] = adata.obs['n_genes_by_counts']
    adata.obs['cell_complexity'] = adata.obs['detected_genes'] / adata.obs['nUMIs']
    print('End calculation of QC metrics.')

    # Original QC plot
    n0 = adata.shape[0]
    print(f'Original cell number: {n0}')

    if doublets is True:
        if doublets_method=='scrublet':
            # Post doublets removal QC plot
            print('Begin of post doublets removal and QC plot')
            rsc.pp.scrublet(adata, random_state=1234,batch_key=batch_key)
            adata_remove = adata[adata.obs['predicted_doublet'], :]
            removed_cells.extend(list(adata_remove.obs_names))
            adata = adata[~adata.obs['predicted_doublet'], :]
            n1 = adata.shape[0]
            print(f'Cells retained after scrublet: {n1}, {n0-n1} removed.')
            print('End of post doublets removal and QC plots.')

        elif doublets_method=='sccomposite':
            adata.obs['sccomposite_doublet']=0
            adata.obs['sccomposite_consistency']=0
            if batch_key is None:
                from ._sccomposite import composite_rna
                multiplet_classification, consistency = composite_rna(adata)
                adata.obs['sccomposite_doublet']=multiplet_classification
                adata.obs['sccomposite_consistency']=consistency
            else:
                for batch in adata.obs[batch_key].unique():
                    from ._sccomposite import composite_rna
                    adata_batch=adata[adata.obs[batch_key]==batch]
                    multiplet_classification, consistency = composite_rna(adata_batch)
                    adata.obs.loc[adata_batch.obs.index,'sccomposite_doublet']=\
                    multiplet_classification
                    adata.obs.loc[adata_batch.obs.index,'sccomposite_consistency']=consistency
            adata_remove = adata[adata.obs['sccomposite_doublet']!=0, :]
            removed_cells.extend(list(adata_remove.obs_names))
            adata = adata[adata.obs['sccomposite_doublet']==0, :]
            n1 = adata.shape[0]
            print(f'Cells retained after sccomposite: {n1}, {n0-n1} removed.')
            print('End of post sccomposite removal and QC plots.')

    # Filters
    print('Filters application (seurat or mads)')
    if mode == 'seurat':
        adata.obs['passing_mt'] = adata.obs['mito_perc'] < tresh['mito_perc']
        adata.obs['passing_nUMIs'] = adata.obs['nUMIs'] > tresh['nUMIs']
        adata.obs['passing_ngenes'] = adata.obs['detected_genes'] > tresh['detected_genes']
    elif mode == 'mads':
        adata.obs['passing_mt'] = adata.obs['mito_perc'] < tresh['mito_perc']
        adata.obs['passing_nUMIs'] = mads_test(adata.obs, 'nUMIs', nmads=nmads, lt=tresh)
        adata.obs['passing_ngenes'] = mads_test(adata.obs, 'detected_genes', nmads=nmads, lt=tresh)
     # Report
    if mode == 'seurat':
        print(f'Lower treshold, nUMIs: {tresh["nUMIs"]}; filtered-out-cells: \
            {n1-np.sum(adata.obs["passing_nUMIs"])}')
        print(f'Lower treshold, n genes: {tresh["detected_genes"]}; \
            filtered-out-cells: {n1-np.sum(adata.obs["passing_ngenes"])}')
        print(f'Lower treshold, mito %: {tresh["mito_perc"]}; \
            filtered-out-cells: {n1-np.sum(adata.obs["passing_mt"])}')
    elif mode == 'mads':
        nUMIs_t = mads(adata.obs, 'nUMIs', nmads=nmads, lt=tresh)
        n_genes_t = mads(adata.obs, 'detected_genes', nmads=nmads, lt=tresh)
        print(f'Tresholds used, nUMIs: ({nUMIs_t[0]}, {nUMIs_t[1]}); filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_nUMIs"])}')
        print(f'Tresholds used, n genes: ({n_genes_t[0]}, {n_genes_t[1]}); filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_ngenes"])}')
        print(f'Lower treshold, mito %: {tresh["mito_perc"]}; filtered-out-cells: \
        {n1-np.sum(adata.obs["passing_mt"])}')
    print(f'Filters applicated.')

    # QC plot
    QC_test = (adata.obs['passing_mt']) & (adata.obs['passing_nUMIs']) & (adata.obs['passing_ngenes'])
    removed = QC_test.loc[lambda x : x is False]
    removed_cells.extend(list(removed.index.values))
    print(f'Total cell filtered out with this last --mode {mode} QC (and its chosen options): \
    {n1-np.sum(QC_test)}')
    adata = adata[QC_test, :]
    n2 = adata.shape[0]
    # Store cleaned adata
    print(f'Cells retained after scrublet and {mode} filtering: {n2}, {n0-n2} removed.')

    # Last gene and cell filter
    rsc.pp.filter_cells(adata,qc_var='detected_genes', min_count=min_genes, \
        max_count=max_genes_ratio*adata.shape[1])
    rsc.pp.filter_genes(adata, min_count=min_cells, max_count=max_cells_ratio*adata.shape[0])

    if adata.uns['status'] is None:
        adata.uns['status'] = {}
    adata.uns['status']['qc']=True
    return adata




def filter_cells(adata: anndata.AnnData,
    min_counts: Optional[int] = None,
    min_genes: Optional[int] = None,
    max_counts: Optional[int] = None,
    max_genes: Optional[int] = None,
    inplace: bool = True,):
    """
    Filter cell outliers based on counts and numbers of genes expressed.
    
    For instance, only keep cells with at least `min_counts` counts or
    `min_genes` genes expressed. This is to filter measurement outliers,
    i.e. “unreliable” observations.
    
    Only provide one of the optional parameters `min_counts`, `min_genes`,
    `max_counts`, `max_genes` per call.
    
    Parameters
    ----------
    data
        The (annotated) data matrix of shape `n_obs` × `n_vars`.
        Rows correspond to cells and columns to genes.
    min_counts
        Minimum number of counts required for a cell to pass filtering.
    min_genes
        Minimum number of genes expressed required for a cell to pass filtering.
    max_counts
        Maximum number of counts required for a cell to pass filtering.
    max_genes
        Maximum number of genes expressed required for a cell to pass filtering.
    inplace
        Perform computation inplace or return result.
    
    Returns
    -------
    Depending on `inplace`, returns the following arrays or directly subsets
    and annotates the data matrix:
    
    cells_subset
        Boolean index mask that does filtering. `True` means that the
        cell is kept. `False` means the cell is removed.
    number_per_cell
        Depending on what was tresholded (`counts` or `genes`),
        the array stores `n_counts` or `n_cells` per gene.
    
    """

    sc.pp.filter_cells(adata, min_genes=min_genes,min_counts=min_counts, max_counts=max_counts,
                       max_genes=max_genes, inplace=inplace)
def filter_genes(adata: anndata.AnnData,
    min_counts: Optional[int] = None,
    min_cells: Optional[int] = None,
    max_counts: Optional[int] = None,
    max_cells: Optional[int] = None,
    inplace: bool = True,):
    """
    Filter genes based on number of cells or counts.
    
    Keep genes that have at least `min_counts` counts or are expressed in at
    least `min_cells` cells or have at most `max_counts` counts or are expressed
    in at most `max_cells` cells.
    
    Only provide one of the optional parameters `min_counts`, `min_cells`,
    `max_counts`, `max_cells` per call.
    
    Parameters
    ----------
    data
        An annotated data matrix of shape `n_obs` × `n_vars`. Rows correspond
        to cells and columns to genes.
    min_counts
        Minimum number of counts required for a gene to pass filtering.
    min_cells
        Minimum number of cells expressed required for a gene to pass filtering.
    max_counts
        Maximum number of counts required for a gene to pass filtering.
    max_cells
        Maximum number of cells expressed required for a gene to pass filtering.
    inplace
        Perform computation inplace or return result.
    
    Returns
    -------
    Depending on `inplace`, returns the following arrays or directly subsets
    and annotates the data matrix
    
    gene_subset
        Boolean index mask that does filtering. `True` means that the
        gene is kept. `False` means the gene is removed.
    number_per_gene
        Depending on what was tresholded (`counts` or `cells`), the array stores
        `n_counts` or `n_cells` per gene.
    
    """

    sc.pp.filter_genes(adata, min_counts=min_counts, min_cells=min_cells, max_counts=max_counts,
                          max_cells=max_cells, inplace=inplace)
    



import networkx as nx
import numpy as np
import pandas as pd
#import cvxpy as cvx

cvx_install = False

def cvxpy():
    global cvx_install
    try:
        import cvxpy as cvx
        cvx_install=True
    except ImportError:
        raise ImportError(
            'Please install the cvxpy: `pip install -U cvxpy`.'
            )


def _root_nodes(directed_graph: nx.DiGraph):
    return set([n for n in directed_graph.nodes() if (directed_graph.in_degree(n) == 0)
                or (list(directed_graph.predecessors(n)) == [n])])


def _end_nodes(directed_graph: nx.DiGraph):
    return set([n for n in directed_graph.nodes() if (directed_graph.out_degree(n) == 0)
                or (list(directed_graph.successors(n)) == [n])])


"""
MDS control for a directed network.
Use efficient graph reduction and then solve the ILP problem using gurobi.
A set S \subset V of nodes in a graph G=(V,E) is a dominating set if every node v \in V is either an element of S 
or adjacent to an element of S.
Refs: [1] Dominating scale-free networks with variable scaling exponent: heterogeneous networks are not difficult to control.
      New Journal of Physics, 2012.
      [2] Critical controllability analysis of directed biological networks using efficient graph reduction.
      Scientific Reports, 2017.
"""


def _MDS_graph_reduction(directed_graph: nx.DiGraph):
    # Critical nodes are driver nodes
    critical_nodes = set()
    redundant_nodes = set()

    # Critical nodes cond. 1: Source nodes are critical (driver) nodes.
    critical_nodes.update(_root_nodes(directed_graph))
    remain_nodes = set(directed_graph.nodes()) - critical_nodes

    noChange = False
    while not noChange:
        noChange = True

        # Critical nodes cond. 2: A node with at least two directed edges to nodes with outdegree 0 and indegree 1.
        in1out0_nodes = set([n for n in directed_graph.nodes() if (directed_graph.in_degree(n) == 1
                                                                   and directed_graph.out_degree(n) == 0)])
        add_critical = set([n for n in list(remain_nodes - in1out0_nodes)
                            if (len(set(directed_graph.successors(n)).intersection(in1out0_nodes)) > 1)])
        if len(add_critical) == 0:
            noChange *= True
        else:
            critical_nodes.update(add_critical)
            remain_nodes = remain_nodes - add_critical
            remove_edges = [(i, n) for n in list(add_critical) for i in directed_graph.predecessors(n)]
            directed_graph.remove_edges_from(remove_edges)
            noChange *= False

        # Redundant nodes cond.: a node with outdegree 0 and has an incoming link from a critical node.
        add_redundant = set([n for n in list(remain_nodes)
                             if (len(set(directed_graph.predecessors(n)).intersection(critical_nodes)) > 0
                                 and (directed_graph.out_degree(n) == 0))])
        if len(add_redundant) == 0:
            noChange *= True
        else:
            redundant_nodes.update(add_redundant)
            remain_nodes = remain_nodes - add_redundant
            directed_graph.remove_nodes_from(add_redundant)
            noChange *= False

    return directed_graph, critical_nodes, redundant_nodes


def MDScontrol(directed_graph: nx.DiGraph, solver='GUROBI'):
    cvxpy()
    global cvx_install
    if cvx_install==True:
        global_imports("cvxpy","cvx")


    print('  Solving MDS problem...')
    directed_graph.remove_edges_from(nx.selfloop_edges(directed_graph))
    reduced_graph = nx.DiGraph(directed_graph)
    intermittent_nodes = set()
    MDS_driver_set = set()

    # Graph reduction
    reduced_graph, critical_nodes, redundant_nodes = _MDS_graph_reduction(reduced_graph)
    print('    {} critical nodes are found.'.format(len(critical_nodes)))

    # Use ILP to find an MDS in the reduced graph
    reduced_graph.remove_nodes_from(list(nx.isolates(reduced_graph)))
    print('    {} nodes left after graph reduction operation.'.format(reduced_graph.number_of_nodes()))
    if reduced_graph.number_of_nodes() == 0:
        print('  {} MDS driver nodes are found.'.format(len(critical_nodes)))
    else:
        print('    Solving the Integer Linear Programming problem on the reduced graph...')
        A = nx.to_numpy_array(reduced_graph)
        A = A + np.eye(A.shape[0]) # A = A + np.diag(np.ones(A.shape[0]))
        # Define the optimization variables
        x = cvx.Variable(reduced_graph.number_of_nodes(), boolean=True)
        # Define the constraints
        constraints = [A @ x >= np.ones(reduced_graph.number_of_nodes())]
        # constraints = [A.T @ x >= np.ones(reduced_graph.number_of_nodes())]
        # constraints = [x[i] + cvx.sum(x[j] for j in range(reduced_graph.number_of_nodes()) if A[i][j]) >= 1 for i in
        #                range(reduced_graph.number_of_nodes())]
        # Define the optimization problem
        obj = cvx.Minimize(cvx.sum(x))
        # Solve the problem
        prob = cvx.Problem(obj, constraints)

        if solver == 'GUROBI':
            # Solve with GUROBI.
            print('      Solving by GUROBI...(', end='')
            prob.solve(solver=cvx.GUROBI, verbose=False)
            print('optimal value with GUROBI:{},'.format(prob.value), end='  ')
        # elif solver == 'XPRESS':
        #        prob.solve(solver=cvx.XPRESS, verbose=False)
        #        print("optimal value with XPRESS:", prob.value)
        else:
            # Solve with SCIP
            print('      Inaccurate solver is selected! Now, solving by SCIP...(', end='')
            prob.solve(solver=cvx.SCIP, verbose=False)
            print('optimal value with SCIP:{},'.format(prob.value), end='  ')
        print('status:{})'.format(prob.status))

        # Set remain nodes that belongs to the MDS as critical nodes
        nodes_idx_map = dict(zip(range(reduced_graph.number_of_nodes()), reduced_graph.nodes()))
        mds_nodes = set([v for k, v in nodes_idx_map.items() if x.value[k] == 1])
        MDS_driver_set = critical_nodes.union(mds_nodes)
        intermittent_nodes = set(reduced_graph.nodes()) - MDS_driver_set
        print('  {} MDS driver genes are found.'.format(len(MDS_driver_set)))

    return MDS_driver_set, intermittent_nodes


"""
FVS control for a directed network.
Use efficient graph reduction and then solve the ILP problem using gurobi. 
A set S \subset V of nodes in a graph G=(V,E) is a feedback vertex set if the removal of these nodes leaves the graph
without feedback loops.
Refs: [1] Structure-based control of complex networks with nonlinear dynamics.
      Proceedings of the National Academy of Sciences, 2017
      [2] On computing the minimum feedback vertex set of a directed graph by contraction operations.
      IEEE Transactions on Computer-Aided Design of Integrated Circuits and Systems, 2000
      [ILP] An exact algorithm for selecting partial scan flip-flops. 1994
"""


def _in0out0(directed_graph: nx.DiGraph):
    remove_set = set([n for n in directed_graph.nodes() if (directed_graph.in_degree(n) == 0
                                                            or directed_graph.out_degree(n) == 0)])
    directed_graph.remove_nodes_from(remove_set)
    return directed_graph, len(remove_set) == 0


def _selfloop(directed_graph: nx.DiGraph, S: set):
    remove_set = list(nx.nodes_with_selfloops(directed_graph))
    S = S.union(set(remove_set))
    directed_graph.remove_nodes_from(remove_set)
    return directed_graph, S, len(remove_set) == 0


def _in1(directed_graph: nx.DiGraph):
    remove_set = set([n for n in directed_graph.nodes() if (directed_graph.in_degree(n) == 1)])
    temp_set = set()
    for u in remove_set:
        v = list(directed_graph.predecessors(u))[0]
        if v != u:
            directed_graph.add_edges_from([(v, w) for w in directed_graph.successors(u)])
            directed_graph.remove_node(u)
        else:
            temp_set.update(u)
    remove_set = remove_set - temp_set
    return directed_graph, len(remove_set) == 0


def _out1(directed_graph: nx.DiGraph):
    remove_set = set([n for n in directed_graph.nodes() if (directed_graph.out_degree(n) == 1)])
    temp_set = set()
    for u in remove_set:
        v = list(directed_graph.successors(u))[0]
        if u != v:
            directed_graph.add_edges_from([(w, v) for w in directed_graph.predecessors(u)])
            directed_graph.remove_node(u)
        else:
            temp_set.update(u)
    remove_set = remove_set - temp_set
    return directed_graph, len(remove_set) == 0


def _PIE(directed_graph: nx.DiGraph):
    G_SCCs = [c for c in nx.strongly_connected_components(directed_graph)]
    edges_SCCs = set()
    for g in G_SCCs:
        edges_SCCs.update(set(directed_graph.subgraph(g).edges()))
    remove_set = set(directed_graph.edges()) - edges_SCCs
    directed_graph.remove_edges_from(list(remove_set))
    return directed_graph, len(remove_set) == 0


def _CORE(directed_graph: nx.DiGraph, nodes_importance: pd.DataFrame, S: set):
    remove_set = set()
    pie = [e for e in directed_graph.edges() if (directed_graph.has_edge(e[1], e[0]) and (e[1] != e[0]))]
    if len(pie) > 0:
        x, y = zip(*pie)
        nodes_pie = set(x).union(set(y))
        piv = [n for n in list(nodes_pie)
               if (set(directed_graph.predecessors(n)) == set(directed_graph.successors(n)))]
        # sort PIV in the ascending order according to the degree and out_degree_importance
        piv_df = pd.DataFrame(directed_graph.degree(piv), index=piv, columns=['id', 'degree'])
        piv_df['importance'] = piv_df.index.map(nodes_importance)
        piv_df.sort_values(by=['importance', 'degree'], ascending=[True, True], inplace=True)
        # piv_df.sort_values(by='degree', ascending=True, inplace=True)
        piv_df['valid'] = True
        for v in piv_df.index:
            if piv_df.loc[v, 'valid']:
                # d-clique
                d_graph = directed_graph.subgraph([v] + list(directed_graph.neighbors(v)))
                is_dclique = True
                for d_i in d_graph:
                    if len(list(nx.bfs_edges(d_graph, d_i, depth_limit=1))) < (len(d_graph) - 1):
                        is_dclique = False

                if is_dclique:
                    S = S.union(set(d_graph.nodes()) - set(v))  # v is CORE
                    remove_set = remove_set.union(set(d_graph.nodes()))
                    piv_df.loc[piv_df.index.isin(d_graph.nodes()), 'valid'] = False
                else:
                    piv_df.loc[v, 'valid'] = False
    directed_graph.remove_nodes_from(remove_set)
    return directed_graph, S, len(remove_set) == 0


def _DOME(directed_graph: nx.DiGraph):
    pie = [e for e in directed_graph.edges() if (directed_graph.has_edge(e[1], e[0]) and (e[1] != e[0]))]
    auxilliary_graph = nx.DiGraph(directed_graph)
    auxilliary_graph.remove_edges_from(pie)
    remove_set = [e for e in auxilliary_graph.edges() if e[0] != e[1] and
                  (set(auxilliary_graph.predecessors(e[0])).issubset(set(auxilliary_graph.predecessors(e[1])))
                   or
                   set(auxilliary_graph.successors(e[1])).issubset(set(auxilliary_graph.successors(e[0]))))]
    directed_graph.remove_edges_from(remove_set)
    return directed_graph, len(remove_set) == 0


def _MFVS_graph_reduction(directed_graph: nx.DiGraph, nodes_importance: pd.DataFrame, S: set):
    all_done = False
    while not all_done:
        all_done = True

        directed_graph, no_change = _in0out0(directed_graph)
        all_done *= no_change

        directed_graph, S, no_change = _selfloop(directed_graph, S)
        all_done *= no_change

        directed_graph, no_change = _out1(directed_graph)
        all_done *= no_change

        directed_graph, no_change = _in1(directed_graph)
        all_done *= no_change

        directed_graph, no_change = _PIE(directed_graph)
        all_done *= no_change

        directed_graph, S, no_change = _CORE(directed_graph, nodes_importance, S)
        all_done *= no_change

        directed_graph, no_change = _DOME(directed_graph)
        all_done *= no_change

    return directed_graph, S


def MFVScontrol(directed_graph: nx.DiGraph, nodes_importance: pd.DataFrame, solver='GUROBI'):
    cvxpy()
    global cvx_install
    if cvx_install==True:
        global_imports("cvxpy","cvx")


    print('  Solving MFVS problem...')
    # Source nodes are critical (driver) nodes
    critical_nodes = set()
    source_nodes = _root_nodes(directed_graph)
    critical_nodes.update(source_nodes)
    reduced_graph = nx.DiGraph(directed_graph)

    while True:
        # Graph reduction
        reduced_graph, critical_nodes = _MFVS_graph_reduction(reduced_graph, nodes_importance, critical_nodes)
        reduced_graph_SCCs = [c for c in nx.strongly_connected_components(reduced_graph)]
        if len(reduced_graph_SCCs) > 1:
            reduced_graph_new = nx.DiGraph()
            for g in reduced_graph_SCCs:
                if len(g) > 1:
                    g = nx.DiGraph(reduced_graph.subgraph(g))
                    g, critical_nodes = _MFVS_graph_reduction(g, nodes_importance, critical_nodes)
                    reduced_graph_new.update(g)
            reduced_graph = nx.DiGraph(reduced_graph_new)

        # Some tricks: If the size of reduced graph is still large, let the node with the largest degree
        # be a critical node, and do graph reduction again.
        if reduced_graph.number_of_nodes() > 150:
            d_seq = pd.merge(pd.DataFrame(reduced_graph.in_degree(), columns=['Node', 'in_Degree']),
                             pd.DataFrame(reduced_graph.out_degree(), columns=['Node', 'out_Degree']),
                             on='Node')
            d_seq['dot_Degree'] = d_seq['in_Degree'] * d_seq['out_Degree']
            node_max = d_seq['Node'][d_seq['dot_Degree'].idxmax()]
            critical_nodes.add(node_max)
            reduced_graph.remove_node(node_max)
        else:
            break
    print('    {} critical nodes are found.'.format(len(critical_nodes)))

    # Use ILP to find an MFVS in the reduced graph
    reduced_graph.remove_nodes_from(list(nx.isolates(reduced_graph)))
    print('    {} nodes left after graph reduction operation.'.format(reduced_graph.number_of_nodes()))
    if reduced_graph.number_of_nodes() == 0:
        print('  {} MFVS driver genes are found.'.format(len(critical_nodes)))
    else:
        print('    Solving the Integer Linear Programming problem on the reduced graph...')
        nodes_idx_map = dict(zip(reduced_graph.nodes(), range(len(reduced_graph.nodes()))))
        # Define the optimization variables
        n = reduced_graph.number_of_nodes()
        x = cvx.Variable(n, boolean=True)
        # Define the constraints
        constraints = []
        w = cvx.Variable(n, integer=True)
        for e in reduced_graph.edges():
            constraints += [w[nodes_idx_map[e[0]]] - w[nodes_idx_map[e[1]]] + n * x[nodes_idx_map[e[0]]] >= 1]
        constraints += [0 <= w, w <= (n - 1)]
        # Define the optimization problem
        obj = cvx.Minimize(cvx.sum(x))
        # Solve the problem
        prob = cvx.Problem(obj, constraints)

        if solver == 'GUROBI':
            # Solve with GUROBI.
            print('      Solving by GUROBI...(', end='')
            prob.solve(solver=cvx.GUROBI, verbose=False)
            print('optimal value with GUROBI:{},'.format(prob.value), end='  ')
        # elif solver=='XPRESS':
        #    prob.solve(solver=cvx.XPRESS, verbose=False)
        #    print("optimal value with XPRESS:", prob.value)
        else:
            # Solve with SCIP
            print('      Inaccurate solver is selected. Now, solving by SCIP...(', end='')
            prob.solve(solver=cvx.SCIP, verbose=False)
            print('optimal value with SCIP:{},'.format(prob.value), end='  ')
        print('status:{})'.format(prob.status))

        # Set remain nodes that belongs to the MDS as critical nodes
        mFVS_remain = set([k for k, v in nodes_idx_map.items() if x.value[v] == 1])
        critical_nodes = critical_nodes.union(mFVS_remain)
        print('  {} MFVS driver nodes are found.'.format(len(critical_nodes)))

    return critical_nodes, source_nodes


def highly_weighted_genes(gene_influence_scores: pd.DataFrame, topK: int = 50):
    v_out = gene_influence_scores.sort_values(by='score_out', ascending=False)
    v_out = v_out.iloc[0:topK, [0]]
    out_critical_genes = set(v_out[v_out > 0].index)

    v_in = gene_influence_scores.sort_values(by='score_in', ascending=False)
    v_in = v_in.iloc[0:topK, [1]]
    in_critical_genes = set(v_in[v_in > 0].index)

    critical_genes = out_critical_genes.union(in_critical_genes)
    return critical_genes, out_critical_genes, in_critical_genes


def driver_regulators(GRN_nx: nx.DiGraph,
                      gene_influence_score: pd.DataFrame,
                      topK: int = 100,
                      driver_union: bool = True,
                      solver: str = 'GUROBI'):
    print('[2] - Identifying driver regulators...')

    # Check if Gurobi can be used successfully
    if solver == 'GUROBI':
        import gurobipy as gp
        gp.Model()

    # MFVS driver nodes
    MFVS_driver_set, source_nodes = MFVScontrol(GRN_nx, gene_influence_score.loc[:, 'score_out'], solver=solver)
    MFVS_driver_set = MFVS_driver_set.union(source_nodes)
    # MDS driver nodes
    MDS_driver_set, MDS_intermittent_nodes = MDScontrol(GRN_nx, solver=solver)
    # Merge two kinds of drivers
    if driver_union:
        driver_set = MDS_driver_set.union(MFVS_driver_set)
    else:
        driver_set = MDS_driver_set.intersection(MFVS_driver_set)

    # Top-ranked genes based on the gene influence score
    critical_genes, out_critical_genes, in_critical_genes = highly_weighted_genes(gene_influence_score, topK//2)

    # The final driver regulators identified by CEFCON
    CEFCON_drivers = driver_set.intersection(critical_genes)

    drivers_df = gene_influence_score.loc[
        list(MFVS_driver_set.union(MDS_driver_set).union(critical_genes)),
        ['influence_score']
    ].copy()
    drivers_df['is_driver_regulator'] = drivers_df.index.isin(list(CEFCON_drivers))
    drivers_df['is_MFVS_driver'] = drivers_df.index.isin(list(MFVS_driver_set))
    drivers_df['is_MDS_driver'] = drivers_df.index.isin(list(MDS_driver_set))
    drivers_df.sort_values(by='influence_score', ascending=False, inplace=True)

    return drivers_df, out_critical_genes, in_critical_genes


def global_imports(modulename,shortname = None, asfunction = False):
    if shortname is None: 
        shortname = modulename
    if asfunction is False:
        globals()[shortname] = __import__(modulename)
    else:        
        globals()[shortname] = __import__(modulename)


from typing import Iterable, Union, Tuple, List
from warnings import warn
import pandas as pd
import numpy as np


import gc

from scipy.sparse import csr_matrix, find, issparse, hstack
from scipy.sparse.linalg import eigs
#import mellon
import scanpy as sc

from .core import run_palantir

from .validation import _validate_obsm_key


class CellNotFoundException(Exception):
    """Exception raised when no valid component is found for the provided cell type."""

    pass


def run_pca(
    data: Union[pd.DataFrame, sc.AnnData],
    n_components: int = 300,
    use_hvg: bool = True,
    pca_key: str = "X_pca",
) -> Union[Tuple[pd.DataFrame, np.array], None]:
    """
    Run PCA on the data.

    Parameters
    ----------
    data : Union[pd.DataFrame, sc.AnnData]
        Dataframe of cells X genes or sc.AnnData object.
        Typically multi-scale space diffusion components.
    n_components : int, optional
        Number of principal components. Default is 300.
    use_hvg : bool, optional
        Whether to use highly variable genes only for PCA. Default is True.
    pca_key : str, optional
        Key to store the PCA projections in obsm of data if it is a sc.AnnData object. Default is 'X_pca'.

    Returns
    -------
    Union[Tuple[pd.DataFrame, np.array], None]
        Tuple of PCA projections of the data and the explained variance.
        If sc.AnnData is passed as data, the results are also written to the input object and None is returned.
    """
    if isinstance(data, pd.DataFrame):
        ad = sc.AnnData(data.values)
    else:
        ad = data
        if pca_key != "X_pca":
            old_pca = ad.obsm.get("X_pca", None)
        else:
            old_pca = None

    # Run PCA
    if not use_hvg:
        n_comps = n_components
    else:
        l_n_comps = min(1000, ad.n_obs - 1, ad.n_vars - 1)
        sc.pp.pca(ad, n_comps=l_n_comps, use_highly_variable=True, zero_center=False)
        try:
            n_comps = np.where(np.cumsum(ad.uns["pca"]["variance_ratio"]) > 0.85)[0][0]
        except IndexError:
            n_comps = n_components

    # Rerun with selection number of components
    n_comps = min(n_comps, ad.n_obs - 1, ad.n_vars - 1)
    sc.pp.pca(ad, n_comps=n_comps, use_highly_variable=use_hvg, zero_center=False)

    if isinstance(data, sc.AnnData):
        data.obsm[pca_key] = ad.obsm["X_pca"]
        if pca_key != "X_pca":
            del data.obsm["X_pca"]
        elif old_pca is not None:
            data.obsm["X_pca"] = old_pca

    pca_projections = pd.DataFrame(ad.obsm[pca_key], index=ad.obs_names)
    return pca_projections, ad.uns["pca"]["variance_ratio"]


def run_low_density_variability(
    ad: sc.AnnData,
    cell_mask: Union[str, np.ndarray, List[str], pd.Series, pd.Index] = "branch_masks",
    density_key: str = "mellon_log_density",
    localvar_key: str = "local_variability",
    score_key: str = "low_density_gene_variability",
) -> np.ndarray:
    """
    Compute the scores aggregated local gene variability in low-density cell-state transitions.

    Parameters
    ----------
    ad : sc.AnnData
        AnnData object containing the gene expression data and pseudotime.
    cell_mask : str, np.ndarray, list of str, pd.Series, pd.Index, optional
        Key to access the mask matrix in the obsm or obs attributes of the AnnData object.
        If cell_mask is a numpy array with shape (ad.n_obs, ), it is used directly.
        If cell_mask is a list of cell names, a pd.Series, or a pd.Index,
        it is used to create a boolean mask of the same length as ad.n_obs.
        Default is 'branch_masks'.
    density_key : str, optional
        Key to access the density values in the obs attribute of the AnnData object.
        Default is 'mellon_log_density'.
    localvar_key : str, optional
        Key to access local variability matrix in the layers of the AnnData object.
        Default is 'local_variability'.
    score_key : str, optional
        Prefix of the key under which the computed scores are stored in the var attribute of the
        AnnData object. Actual keys are '{score_key}_{branch_name}' if cell_mask points to an ad.obsm.
        Default is 'low_density_gene_variability'.

    Returns
    -------
    low_density_scores : np.ndarray
        A numpy array of scores for each gene.

    Raises
    ------
    ValueError
        If any of the provided keys are not found in the appropriate fields of the AnnData object.
    """
    if density_key not in ad.obs.columns:
        raise ValueError(f"'{density_key}' not found in ad.obs.")
    dip_weight = np.exp(-ad.obs[density_key].values)
    if localvar_key not in ad.layers.keys():
        raise ValueError(f"'{localvar_key}' not found in ad.layers.")
    local_var = ad.layers[localvar_key]
    if isinstance(cell_mask, str):
        if cell_mask in ad.obsm.keys():
            masks, branch_names = _validate_obsm_key(ad, cell_mask, as_df=False)
            if masks.ndim == 1:
                masks = masks[:, None]
        elif cell_mask in ad.obs.columns:
            branch_names = [
                "_" + cell_mask,
            ]
            masks = ad.obs[cell_mask].values[:, None]
        else:
            raise ValueError(f"'{cell_mask}' not found in ad.obsm or ad.obs.columns.")
    elif isinstance(cell_mask, np.ndarray) and cell_mask.shape == (ad.n_obs,):
        masks = cell_mask[:, None]
        branch_names = [
            "",
        ]
    elif isinstance(cell_mask, (list, pd.Series, pd.Index)):
        masks = ad.obs_names.isin(cell_mask)[:, None]
        branch_names = [
            "",
        ]
    else:
        raise ValueError(
            "cell_mask must be either a string key, a numpy array with shape "
            "(ad.n_obs, ), a list of cell names, a pd.Series, or a pd.Index."
        )

    out_columns = list()
    for i, branch in enumerate(branch_names):
        idx = masks[:, i]
        colname = score_key + "_" + branch
        ad.var[colname] = np.mean(
            local_var[idx, :] * dip_weight[idx, None],
            axis=0,
        )
        out_columns.append(colname)

    return ad.var[out_columns].values


def run_density(
    ad: sc.AnnData,
    repr_key: str = "DM_EigenVectors",
    density_key: str = "mellon_log_density",
    **kwargs,
) -> np.ndarray:
    """
    Compute cell-state density with Mellon.

    This function uses the Mellon algorithm to compute the density of cell states, which is stored in the obs attribute of the
    AnnData object. The function returns the computed density. If 'DM_EigenVectors' is not found in the AnnData object,
    an error is raised suggesting the user to run the function `palantir.utils.run_diffusion_maps(ad)`.

    Additionally, the density prediction model is serialized and stored in the `.uns` attribute of the AnnData object
    under the key `'{density_key}_predictor'`. This can be deserialized and used for prediction by using the `mellon.Predictor.from_dict()`
    method.

    Parameters
    ----------
    ad : sc.AnnData
        AnnData object containing the gene expression data and pseudotime.
    repr_key : str, optional
        Key to retrieve cell-state representation from the sc.AnnData object. Default is 'DM_EigenVectors'.
    density_key : str, optional
        Key under which the computed density values are stored in the obs of the AnnData object.
        Default is 'mellon_log_density'.
    **kwargs : dict
        Additional keyword arguments to be passed to mellon.DensityEstimator.

    Returns
    -------
    log_density : np.ndarray
        A numpy array of log density values computed for each cell.

    Raises
    ------
    ValueError
        If `repr_key` is not found in `ad.obsm`.
    """
    if repr_key not in ad.obsm:
        raise ValueError(
            f"'{repr_key}' not found in ad.obsm. "
            "Run `palantir.utils.run_diffusion_maps(ad)` to "
            "compute diffusion map eigenvectors."
        )
    X = ad.obsm[repr_key]

    # Set the default arguments for mellon.DensityEstimator
    mellon_args = dict()
    mellon_args.update(kwargs)
    import mellon
    dest = mellon.DensityEstimator(**mellon_args)
    log_density = np.asarray(dest.fit_predict(X))

    ad.obs[density_key] = log_density
    ad.obs[density_key + "_clipped"] = np.clip(
        log_density, *np.quantile(log_density, [0.01, 1])
    )
    ad.uns[density_key + "_predictor"] = dest.predict.to_dict()

    return log_density


def run_density_evaluation(
    in_ad: sc.AnnData,
    out_ad: sc.AnnData,
    predictor_key: str = "mellon_log_density_predictor",
    repr_key: str = "DM_EigenVectors",
    density_key: str = "cross_log_density",
    **kwargs,
) -> np.ndarray:
    """
    Evaluates the density function of `in_ad.uns[predictor_key]` on the representations of `out_ad.obsm[repr_key]`.

    Parameters
    ----------
    in_ad : sc.AnnData
        AnnData object containing the gene expression data and the serialized predictor.
    out_ad : sc.AnnData
        AnnData object containing the gene expression data and representations to be evaluated.
    predictor_key : str, optional
        Key to access the predictor in the uns of the `in_ad` AnnData object. Default is 'mellon_log_density_predictor'.
    repr_key : str, optional
        Key to access representations in the obsm of the `out_ad` AnnData object. Default is 'DM_EigenVectors'.
    density_key : str, optional
        Key under which the computed density values are stored in the obs of the `out_ad` AnnData object.
        Default is 'cross_log_density'.
    **kwargs : dict
        Additional keyword arguments, unused in this function.

    Returns
    -------
    log_density : np.ndarray
        A numpy array of log density values computed for each cell in `out_ad`.

    Raises
    ------
    ValueError
        If `repr_key` is not found in `out_ad.obsm` or `predictor_key` is not found in `in_ad.uns`.
    """
    if repr_key not in out_ad.obsm:
        raise ValueError(
            f"'{repr_key}' not found in out_ad.obsm. "
            "Run `palantir.utils.run_diffusion_maps(ad)` to "
            "compute diffusion map eigenvectors."
        )
    if predictor_key not in in_ad.uns:
        raise ValueError(
            f"'{predictor_key}' not found in in_ad.uns. "
            "Run `run_density(ad)` to compute the density estimator."
        )

    X = out_ad.obsm[repr_key]
    import mellon
    predictor = mellon.Predictor.from_dict(in_ad.uns[predictor_key])
    log_density = predictor(X)
    out_ad.obs[density_key] = log_density
    out_ad.obs[density_key + "_clipped"] = np.clip(
        log_density, *np.quantile(log_density, [0.01, 1])
    )

    return log_density


def compute_kernel(
    data: Union[pd.DataFrame, sc.AnnData],
    knn: int = 30,
    alpha: float = 0,
    pca_key: str = "X_pca",
    kernel_key: str = "DM_Kernel",
) -> csr_matrix:
    """
    Compute the adaptive anisotropic diffusion kernel.

    Parameters
    ----------
    data : Union[pd.DataFrame, sc.AnnData]
        Data points (rows) in a feature space (columns) for pd.DataFrame.
        For sc.AnnData, it uses the .X attribute.
    knn : int
        Number of nearest neighbors for adaptive kernel calculation. Default is 30.
    alpha : float
        Normalization parameter for the diffusion operator. Default is 0.
    pca_key : str, optional
        Key to retrieve PCA projections from data if it is a sc.AnnData object. Default is 'X_pca'.
    kernel_key : str, optional
        Key to store the kernel in obsp of data if it is a sc.AnnData object. Default is 'DM_Kernel'.

    Returns
    -------
    csr_matrix
        Computed kernel matrix.
    """

    # If the input is sc.AnnData, convert it to a DataFrame
    if isinstance(data, sc.AnnData):
        data_df = pd.DataFrame(data.obsm[pca_key], index=data.obs_names)
    else:
        data_df = data

    N = data_df.shape[0]
    temp = sc.AnnData(data_df.values)
    sc.pp.neighbors(temp, n_pcs=0, n_neighbors=knn)
    kNN = temp.obsp["distances"]

    adaptive_k = int(np.floor(knn / 3))
    adaptive_std = np.zeros(N)
    for i in np.arange(N):
        adaptive_std[i] = np.sort(kNN.data[kNN.indptr[i] : kNN.indptr[i + 1]])[
            adaptive_k - 1
        ]

    x, y, dists = find(kNN)
    dists /= adaptive_std[x]
    W = csr_matrix((np.exp(-dists), (x, y)), shape=[N, N])

    kernel = W + W.T

    if alpha > 0:
        D = np.ravel(kernel.sum(axis=1))
        D[D != 0] = D[D != 0] ** (-alpha)
        mat = csr_matrix((D, (range(N), range(N))), shape=[N, N])
        kernel = mat.dot(kernel).dot(mat)

    if isinstance(data, sc.AnnData):
        data.obsp[kernel_key] = kernel

    return kernel


def diffusion_maps_from_kernel(
    kernel: csr_matrix, n_components: int = 10, seed: Union[int, None] = 0
):
    """
    Compute the diffusion map given a kernel matrix.

    Parameters
    ----------
    kernel : csr_matrix
        Precomputed kernel matrix.
    n_components : int
        Number of diffusion components to compute. Default is 10.
    seed : Union[int, None]
        Seed for random initialization. Default is 0.

    Returns
    -------
    dict
        T-matrix (T), Diffusion components (EigenVectors) and corresponding eigenvalues (EigenValues).
    """
    N = kernel.shape[0]
    D = np.ravel(kernel.sum(axis=1))
    D[D != 0] = 1 / D[D != 0]
    T = csr_matrix((D, (range(N), range(N))), shape=[N, N]).dot(kernel)

    np.random.seed(seed)
    v0 = np.random.rand(min(T.shape))
    D, V = eigs(T, n_components, tol=1e-4, maxiter=1000, v0=v0)

    D = np.real(D)
    V = np.real(V)
    inds = np.argsort(D)[::-1]
    D = D[inds]
    V = V[:, inds]

    for i in range(V.shape[1]):
        V[:, i] = V[:, i] / np.linalg.norm(V[:, i])

    return {"T": T, "EigenVectors": pd.DataFrame(V), "EigenValues": pd.Series(D)}


def run_diffusion_maps(
    data: Union[pd.DataFrame, sc.AnnData],
    n_components: int = 10,
    knn: int = 30,
    alpha: float = 0,
    seed: Union[int, None] = 0,
    pca_key: str = "X_pca",
    kernel_key: str = "DM_Kernel",
    sim_key: str = "DM_Similarity",
    eigval_key: str = "DM_EigenValues",
    eigvec_key: str = "DM_EigenVectors",
):
    """
    Run Diffusion maps using the adaptive anisotropic kernel.

    Parameters
    ----------
    data : Union[pd.DataFrame, sc.AnnData]
        PCA projections of the data or adjacency matrix.
        If sc.AnnData is passed, its obsm[pca_key] is used and the result is written to
        its obsp[kernel_key], obsm[eigvec_key], and uns[eigval_key].
    n_components : int, optional
        Number of diffusion components. Default is 10.
    knn : int, optional
        Number of nearest neighbors for graph construction. Default is 30.
    alpha : float, optional
        Normalization parameter for the diffusion operator. Default is 0.
    seed : Union[int, None], optional
        Numpy random seed, randomized if None, set to an arbitrary integer for reproducibility.
        Default is 0.
    pca_key : str, optional
        Key to retrieve PCA projections from data if it is a sc.AnnData object. Default is 'X_pca'.
    kernel_key : str, optional
        Key to store the kernel in obsp of data if it is a sc.AnnData object. Default is 'DM_Kernel'.
    sim_key : str, optional
        Key to store the similarity in obsp of data if it is a sc.AnnData object. Default is 'DM_Similarity'.
    eigval_key : str, optional
        Key to store the EigenValues in uns of data if it is a sc.AnnData object. Default is 'DM_EigenValues'.
    eigvec_key : str, optional
        Key to store the EigenVectors in obsm of data if it is a sc.AnnData object. Default is 'DM_EigenVectors'.

    Returns
    -------
    dict
        Diffusion components, corresponding eigen values and the diffusion operator.
        If sc.AnnData is passed as data, these results are also written to the input object
        and returned.
    """

    if isinstance(data, sc.AnnData):
        data_df = pd.DataFrame(data.obsm[pca_key], index=data.obs_names)
    else:
        data_df = data

    if not isinstance(data_df, pd.DataFrame) and not issparse(data_df):
        raise ValueError("'data_df' should be a pd.DataFrame or sc.AnnData")

    if not issparse(data_df):
        kernel = compute_kernel(data_df, knn, alpha)
    else:
        warn(
            "'data' is a sparse matrix and will be interpreted as kernel. "
            "To avoid this warning compute diffusion maps from a precompued kernel using "
            "palantir.utils.diffusion_maps_from_kernel()."
        )
        kernel = data_df

    res = diffusion_maps_from_kernel(kernel, n_components, seed)

    res["kernel"] = kernel
    if not issparse(data_df):
        res["EigenVectors"].index = data_df.index

    if isinstance(data, sc.AnnData):
        data.obsp[kernel_key] = res["kernel"]
        data.obsp[sim_key] = res["T"]
        data.obsm[eigvec_key] = res["EigenVectors"].values
        data.uns[eigval_key] = res["EigenValues"].values

    return res


def _dot_helper_func(x, y):
    return x.dot(y)


def _local_var_helper(expressions, distances):
    if hasattr(expressions, "todense"):

        def cast(x):
            return x.todense()

        issparse = True
    else:

        def cast(x):
            return x

        issparse = False
    from tqdm import trange
    for cell in trange(expressions.shape[0]):
        neighbors = distances.getrow(cell).indices if issparse else slice(None)
        try:
            neighbor_expression = cast(expressions[neighbors, :])
            cell_expression = cast(expressions[cell, :])
            expr_deltas = np.array(neighbor_expression - cell_expression)
        except ValueError:
            raise ValueError(f"This cell caused the error: {cell}")
        expr_distance = np.sqrt(np.sum(expr_deltas**2, axis=1, keepdims=True))
        change_rate = expr_deltas / (expr_distance+0.000001)
        yield np.max(change_rate**2, axis=0)



def _local_var_helper_(expressions, distances):
    import numpy as np
    from tqdm import tqdm
    from numba import njit
    # 检查是否为稀疏矩
    is_sparse = hasattr(expressions, "todense")
    cast = (lambda x: x.todense()) if is_sparse else (lambda x: x)

    # 获取总行数
    num_cells = expressions.shape[0]

    @njit
    def compute_change_rate(neighbor_expression, cell_expression):
        expr_deltas = neighbor_expression - cell_expression
        expr_distance = np.sqrt(np.sum(expr_deltas ** 2, axis=1, keepdims=True))
        change_rate = expr_deltas / expr_distance
        return np.max(change_rate ** 2, axis=0)
    
    # 初始化结果数组
    results = np.zeros((num_cells, expressions.shape[1]))

    for cell in tqdm(range(num_cells), desc="Processing cells"):
        if is_sparse:
            neighbors = distances.getrow(cell).indices
        else:
            neighbors = np.arange(num_cells)
        
        try:
            neighbor_expression = cast(expressions[neighbors, :])
            cell_expression = cast(expressions[cell, :])
        except ValueError:
            raise ValueError(f"This cell caused the error: {cell}")
        
        results[cell] = compute_change_rate(neighbor_expression, cell_expression)
    
    return results

# 使用方法：例如
# expressions 和 distances 需要在调用时定义
# results = _local_var_helper(expressions, distances)


def run_local_variability(
    ad: sc.AnnData,
    expression_key: str = "MAGIC_imputed_data",
    distances_key: str = "distances",
    localvar_key: str = "local_variability",
) -> np.ndarray:
    """
    Compute local gene variability scores for each cell.

    This function calculates the variability in gene expression in a local neighbourhood for each cell.
    It adds the result to the layers of the given AnnData object under the specified key.

    Parameters
    ----------
    ad : sc.AnnData
        AnnData object containing the gene expression data and pseudotime.
    expression_key : str, optional
        Key to access the gene expression data in the layers of the AnnData object.
        If None, uses raw expression data in .X. Default is 'MAGIC_imputed_data'.
    distances_key : str, optional
        Key to access the distances matrix in the obsm of the AnnData object.
        Default is 'distances'.
    localvar_key : str, optional
        Key under which the computed local variability matrix is stored in the layers of the AnnData object.
        Default is 'local_variability'.

    Returns
    -------
    local_variability : np.ndarray
        A 2D numpy array of local variability scores for each gene in each cell.
    """

    if expression_key:
        if expression_key not in ad.layers:
            raise KeyError(f"'{expression_key}' not found in .layers.")
        X = ad.layers[expression_key]
    else:
        X = ad.X

    from scipy.sparse import issparse, csr_matrix
    if issparse(X):
        pass
    else:
        print("Converting to sparse matrix")
        X=csr_matrix(X)

    if distances_key not in ad.obsp:
        raise KeyError(f"'{distances_key}' not found in .obsp.")
    X_dists = ad.obsp[distances_key]

    local_variability = np.stack(list(_local_var_helper(X, X_dists)))

    ad.layers[localvar_key] = local_variability

    return local_variability


def run_magic_imputation(
    data: Union[np.ndarray, pd.DataFrame, sc.AnnData, csr_matrix],
    dm_res: Union[dict, None] = None,
    n_steps: int = 3,
    sim_key: str = "DM_Similarity",
    expression_key: str = None,
    imputation_key: str = "MAGIC_imputed_data",
    n_jobs: int = -1,
) -> Union[pd.DataFrame, None, csr_matrix]:
    """
    Run MAGIC imputation on the data.

    Parameters
    ----------
    data : Union[np.ndarray, pd.DataFrame, sc.AnnData, csr_matrix]
        Array or DataFrame of cells X genes, sc.AnnData object, or a sparse csr_matrix.
    dm_res : Union[dict, None], optional
        Diffusion map results from run_diffusion_maps.
        If None and data is a sc.AnnData object, its obsp[kernel_key] is used. Default is None.
    n_steps : int, optional
        Number of steps in the diffusion operator. Default is 3.
    expression_key : str, optional
        Key to access the gene expression data in the layers of the AnnData object.
        If None, uses raw expression data in .X. Default is None.
    sim_key : str, optional
        Key to access the similarity in obsp of data if it is a sc.AnnData object.
        Default is 'DM_Similarity'.
    imputation_key : str, optional
        Key to store the imputed data in layers of data if it is a sc.AnnData object. Default is 'MAGIC_imputed_data'.
    n_jobs : int, optional
        Number of cores to use for parallel processing. If -1, all available cores are used. Default is -1.

    Returns
    -------
    Union[np.ndarray, pd.DataFrame, None, csr_matrix]
        Imputed data matrix. If sc.AnnData is passed as data, the result is written to its layers[imputation_key].
    """
    from joblib import Parallel, delayed
    if isinstance(data, sc.AnnData):
        if expression_key is not None:
            if expression_key not in data.layers.keys():
                raise ValueError(
                    f"expression_key '{expression_key}' not found in .layers."
                )
            X = data.layers[expression_key]
        else:
            X = data.X
        if dm_res is None:
            T = data.obsp[sim_key]
    elif isinstance(data, pd.DataFrame):
        X = data.values
    elif issparse(data):  # assuming csr_matrix
        X = data
    else:  # assuming np.ndarray
        X = data

    if dm_res is not None:
        T = dm_res["T"]
    elif not isinstance(data, sc.AnnData):
        raise ValueError(
            "Diffusion map results (dm_res) must be provided if data is not sc.AnnData"
        )

    # Preparing the operator
    T_steps = (T**n_steps).astype(np.float32)

    # Define chunks of columns for parallel processing
    chunks = np.append(np.arange(0, X.shape[1], 100), [X.shape[1]])

    # Run the dot product in parallel on chunks
    res = Parallel(n_jobs=n_jobs)(
        delayed(_dot_helper_func)(T_steps, X[:, chunks[i - 1] : chunks[i]])
        for i in range(1, len(chunks))
    )

    # Stack the results together
    if issparse(X):
        imputed_data = hstack(res).todense()
    else:
        imputed_data = np.hstack(res)

    # Set small values to zero
    imputed_data[imputed_data < 1e-2] = 0

    # Clean up
    gc.collect()

    if isinstance(data, sc.AnnData):
        data.layers[imputation_key] = np.asarray(imputed_data)

    if isinstance(data, pd.DataFrame):
        imputed_data = pd.DataFrame(
            imputed_data, index=data.index, columns=data.columns
        )

    return imputed_data


def determine_multiscale_space(
    dm_res: Union[dict, sc.AnnData],
    n_eigs: Union[int, None] = None,
    eigval_key: str = "DM_EigenValues",
    eigvec_key: str = "DM_EigenVectors",
    out_key: str = "DM_EigenVectors_multiscaled",
) -> Union[pd.DataFrame, None]:
    """
    Determine the multi-scale space of the data.

    Parameters
    ----------
    dm_res : Union[dict, sc.AnnData]
        Diffusion map results from run_diffusion_maps.
        If sc.AnnData is passed, its uns[eigval_key] and obsm[eigvec_key] are used.
    n_eigs : Union[int, None], optional
        Number of eigen vectors to use. If None is specified, the number
        of eigen vectors will be determined using the eigen gap. Default is None.
    eigval_key : str, optional
        Key to retrieve EigenValues from dm_res if it is a sc.AnnData object. Default is 'DM_EigenValues'.
    eigvec_key : str, optional
        Key to retrieve EigenVectors from dm_res if it is a sc.AnnData object. Default is 'DM_EigenVectors'.
    out_key : str, optional
        Key to store the result in obsm of dm_res if it is a sc.AnnData object. Default is 'DM_EigenVectors_multiscaled'.

    Returns
    -------
    Union[pd.DataFrame, None]
        Multi-scale data matrix. If sc.AnnData is passed as dm_res, the result
        is written to its obsm[out_key] and None is returned.
    """
    if isinstance(dm_res, sc.AnnData):
        eigenvectors = dm_res.obsm[eigvec_key]
        if not isinstance(eigenvectors, pd.DataFrame):
            eigenvectors = pd.DataFrame(eigenvectors, index=dm_res.obs_names)
        dm_res_dict = {
            "EigenValues": dm_res.uns[eigval_key],
            "EigenVectors": eigenvectors,
        }
    else:
        dm_res_dict = dm_res

    if not isinstance(dm_res_dict, dict):
        raise ValueError("'dm_res' should be a dict or a sc.AnnData instance")
    if n_eigs is None:
        vals = np.ravel(dm_res_dict["EigenValues"])
        n_eigs = np.argsort(vals[: (len(vals) - 1)] - vals[1:])[-1] + 1
        if n_eigs < 3:
            n_eigs = np.argsort(vals[: (len(vals) - 1)] - vals[1:])[-2] + 1

    # Scale the data
    use_eigs = list(range(1, n_eigs))
    eig_vals = np.ravel(dm_res_dict["EigenValues"][use_eigs])
    data = dm_res_dict["EigenVectors"].values[:, use_eigs] * (eig_vals / (1 - eig_vals))
    data = pd.DataFrame(data, index=dm_res_dict["EigenVectors"].index)

    if isinstance(dm_res, sc.AnnData):
        dm_res.obsm[out_key] = data.values

    return data


def _return_cell(ec, obs_names, celltype, mm, dcomp):
    """
    Helper function to print and return the early cell.

    Args:
        ec (int): Index of the early cell.
        obs_names (list): Names of cells.
        celltype (str): The cell type of interest.
        mm (str): Max/min status of the diffusion component.
        dcomp (int): Index of diffusion component.

    Returns:
        str: Name of the early cell.
    """
    early_cell = obs_names[ec]
    print(
        f"Using {early_cell} for cell type {celltype} which is {mm} in "
        f"diffusion component {dcomp}."
    )
    return early_cell


def early_cell(
    ad: sc.AnnData,
    celltype: str,
    celltype_column: str = "celltype",
    eigvec_key: str = "DM_EigenVectors_multiscaled",
    fallback_seed: int = None,
):
    """
    Helper function to determine 'early_cell' for 'run_palantir'.
    It identifies the cell of 'celltype' at the extremes of the state space represented by diffusion maps.

    Parameters
    ----------
    ad : sc.AnnData
        Annotated data matrix.
    celltype : str
        The specific cell type of interest for determining the early cell.
    celltype_column : str, optional
        Name of the column in the obs of the Anndata object where the cell type information is stored.
        Default is 'celltype'.
    eigvec_key : str, optional
        Key to access multiscale space diffusion components from obsm of ad.
        Default is 'DM_EigenVectors_multiscaled'.
    fallback_seed : int, optional
        Seed for random number generator in fallback method. If not specified, no seed is used.
        Default is None.

    Returns
    -------
    str
        Name of the early cell for the given cell type.

    Raises
    ------
    CellNotFoundException
        If no valid cell of the specified type can be found at the extremes of the diffusion map.
    """
    if not isinstance(ad, sc.AnnData):
        raise ValueError("'ad' should be an instance of sc.AnnData")

    if eigvec_key not in ad.obsm:
        raise ValueError(
            f"'{eigvec_key}' not found in ad.obsm. "
            "Run `palantir.utils.run_diffusion_maps(ad)` to "
            "compute diffusion map eigenvectors."
        )
    eigenvectors = ad.obsm[eigvec_key]
    if isinstance(eigenvectors, pd.DataFrame):
        eigenvectors = eigenvectors.values

    if not isinstance(celltype_column, str):
        raise ValueError(f"celltype_column='{celltype_column}' should be a string")

    if celltype_column not in ad.obs.columns:
        raise ValueError(
            f"celltype_column='{celltype_column}' should be a column of ad.obs."
        )

    if not isinstance(celltype, str):
        raise ValueError("celltype should be a string")

    if celltype not in ad.obs[celltype_column].values:
        raise ValueError(
            f"Celltype '{celltype}' not found in ad.obs['{celltype_column}']."
        )

    if fallback_seed is not None and not isinstance(fallback_seed, int):
        raise ValueError("'fallback_seed' should be an integer")

    for dcomp in range(eigenvectors.shape[1]):
        ec = eigenvectors[:, dcomp].argmax()
        if ad.obs[celltype_column][ec] == celltype:
            return _return_cell(ec, ad.obs_names, celltype, "max", dcomp)
        ec = eigenvectors[:, dcomp].argmin()
        if ad.obs[celltype_column][ec] == celltype:
            return _return_cell(ec, ad.obs_names, celltype, "min", dcomp)

    if fallback_seed is not None:
        print("Falling back to slow early cell detection.")
        return fallback_terminal_cell(
            ad, celltype, celltype_column=celltype_column, seed=fallback_seed
        )

    raise CellNotFoundException(
        f"No valid component found: {celltype} "
        "Consider increasing the number of diffusion components "
        "('n_components' in palantir.utils.run_diffusion_maps) "
        "or specify a 'fallback_seed' to determine an early cell based on "
        f"reverse pseudotime starting from random non-{celltype} cell."
    )


def fallback_terminal_cell(
    ad: sc.AnnData,
    celltype: str,
    celltype_column: str = "anno",
    eigvec_key: str = "DM_EigenVectors_multiscaled",
    seed: int = 2353,
):
    """
    Fallback method to identify terminal cells when no valid diffusion component
    is found for the specified cell type.

    Parameters
    ----------
    ad : sc.AnnData
        Annotated data matrix.
    celltype : str
        The specific cell type of interest for determining the terminal cell.
    celltype_column : str, optional
        Name of the column in the obs of the Anndata object where the cell type information is stored.
        Default is 'anno'.
    eigvec_key : str, optional
        Key to access multiscale space diffusion components from obsm of ad.
        Default is 'DM_EigenVectors_multiscaled'.
    seed : int, optional
        Seed for random number generator in fallback method. If not specified, no seed is used.
        Default is 2353.

    Returns
    -------
    str
        Name of the terminal cell for the given cell type.

    """
    other_cells = ad.obs_names[ad.obs[celltype_column] != celltype]
    fake_early_cell = other_cells.to_series().sample(1, random_state=seed)[0]
    pr_res = run_palantir(
        ad,
        fake_early_cell,
        eigvec_key=eigvec_key,
        terminal_states=None,
        use_early_cell_as_start=True,
    )
    idx = ad.obs[celltype_column] == celltype
    ec = pr_res.pseudotime[idx].argmax()
    early_cell = ad.obs_names[idx][ec]
    print(
        f"Using {early_cell} for cell type {celltype} which is latest cell in "
        f"{celltype} when starting from {fake_early_cell}."
    )
    return early_cell


def find_terminal_states(
    ad: sc.AnnData,
    celltypes: Iterable,
    celltype_column: str = "celltype",
    eigvec_key: str = "DM_EigenVectors_multiscaled",
    fallback_seed: int = None,
):
    """
    Identifies terminal states for a list of cell types in the AnnData object.

    This function iterates over the provided cell types, trying to find a terminal cell for each one
    using the 'early_cell' function. If no valid component is found for a cell type, it emits a warning and
    proceeds to the next cell type.

    Parameters
    ----------
    ad : sc.AnnData
        Annotated data matrix from Scanpy. It should contain computed diffusion maps.
    celltypes : Iterable
        An iterable such as a list or tuple of cell type names for which terminal states should be identified.
    celltype_column : str, optional
        The name of the column in the obs dataframe of the Anndata object where the cell type information is
        stored. By default, it is 'celltype'.
    eigvec_key : str, optional
        Key to access multiscale space diffusion components from obsm of ad.
        Default is 'DM_EigenVectors_multiscaled'.
    fallback_seed : int, optional
        Seed for the random number generator used in the fallback method. Defaults to None, in which case
        the random number generator will be randomly seeded.

    Returns
    -------
    pd.Series
        A pandas Series where the indices are the cell types and the values are the names of the terminal cells.
        If no terminal cell is found for a cell type, it will not be included in the series.
    """
    terminal_states = pd.Series(dtype=str)
    for ct in celltypes:
        try:
            cell = early_cell(ad, ct, celltype_column, eigvec_key, fallback_seed)
        except CellNotFoundException:
            warn(
                f"No valid component found: {ct} "
                "Consider increasing the number of diffusion components "
                "('n_components' in palantir.utils.run_diffusion_maps). "
                f"The cell type {ct} will be skipped."
            )
            continue
        terminal_states[cell] = ct
    return terminal_states



"""Module providing a encapsulation of spatrio."""
from ..externel.STT import tl,pl
from typing import Any
import scanpy as sc
import numpy as np
import pandas as pd
from .._settings import add_reference

class STT(object):
    """Class representing the object of STT."""
    def __init__(self,adata,spatial_loc='xy_loc',region='Region'):
        self.adata=adata
        self.adata_aggr=None
        self.spatial_loc=spatial_loc
        self.region=region
        if 'attractor' not in self.adata.obs.keys():
            self.adata.obs['attractor'] = self.adata.obs[region]

    def stage_estimate(self):
        """
        Estimate the stage of the cells based on the joint clustering of spliced and unspliced data.
        """
        u = self.adata.layers['unspliced']
        s = self.adata.layers['spliced']
        if 'toarray' in dir(u):
            u = u.toarray()
            s = s.toarray()
        x_all = np.concatenate((u,s),axis = 1)
        adata_aggr = sc.AnnData(X=x_all)
        sc.tl.pca(adata_aggr, svd_solver='arpack')
        sc.pp.neighbors(adata_aggr)
        sc.tl.leiden(adata_aggr,resolution = 0.15)
        self.adata.obs['joint_leiden'] = adata_aggr.obs['leiden'].values
        print(f"...estimate stage: {len(self.adata.obs['joint_leiden'].unique())}")
        add_reference(self.adata,'STT','spatial transition tensor with STT')

    def train(self,
            n_states: int = 9,
            n_iter: int = 15,
            weight_connectivities: float = 0.5,
            n_neighbors: int = 50,
            thresh_ms_gene: float = 0.2,
            spa_weight: float = 0.3,
            **kwargs: Any
        ) -> None:
        """
        Train the model.
        """
        self.adata_aggr = tl.dynamical_iteration(self.adata,n_states = n_states,
                                    n_iter = n_iter, weight_connectivities = weight_connectivities,
                                    n_neighbors = n_neighbors,thresh_ms_gene = thresh_ms_gene,
                                    spa_weight =spa_weight,**kwargs)
        self.adata.obsm[f'X_{self.spatial_loc}'] = self.adata.obsm[self.spatial_loc]
        self.adata_aggr.obsm[f'X_{self.spatial_loc}']=self.adata.obsm[self.spatial_loc]
        self.adata_aggr.obsm[f'X_{self.spatial_loc}_aggr']=self.adata.obsm[self.spatial_loc]
        self.adata.obsm[f'X_{self.spatial_loc}_aggr']=self.adata.obsm[self.spatial_loc]
        add_reference(self.adata,'STT','spatial transition tensor with STT')

    def load(self,adata,adata_aggr):
        """
        define the adata and adata_aggr
        """
        self.adata=adata
        self.adata_aggr=adata_aggr
        add_reference(self.adata,'STT','spatial transition tensor with STT')

    def compute_pathway(self,pathway_dict):
        """
        Compute the pathway of the cells
        """
        return tl.compute_pathway(self.adata,self.adata_aggr,pathway_dict)

    def plot_pathway(self,label_fontsize=20,**kwargs):
        """
        Plot the pathway of the cells
        """
        fig = pl.plot_pathway(self.adata,**kwargs)
        for ax in fig.axes:
            ax.set_xlabel('Embedding 1', fontsize=label_fontsize)  # Adjust font size as needed
            ax.set_ylabel('Embedding 2', fontsize=label_fontsize)  # Adjust font size as needed
        return fig

    def plot_tensor_pathway(self,pathway_name,**kwargs):
        """
        Plot the tensor pathway of the cells
        """
        ax=pl.plot_tensor_pathway(self.adata,self.adata_aggr,
                                  pathway_name = pathway_name,**kwargs)
        return ax

    def plot_tensor(self,list_attractor,**kwargs):
        """
        Plot the tensor of the cells
        """
        return pl.plot_tensor(self.adata, self.adata_aggr, 
                          list_attractor = list_attractor,basis = self.spatial_loc,**kwargs)

    def construct_landscape(self,coord_key = 'X_xy_loc',**kwargs):
        """
        Construct the landscape of the cells
        """
        tl.construct_landscape(self.adata, coord_key = coord_key,**kwargs)
        self.adata.obsm['trans_coord'] = self.adata.uns['land_out']['trans_coord']

    def infer_lineage(self,**kwargs):
        """
        Infer the lineage of the cells
        """
        return pl.infer_lineage(self.adata,**kwargs)

    def plot_landscape(self,**kwargs):
        """
        Plot the landscape of the cells
        """
        return pl.plot_landscape(self.adata,**kwargs)

    def plot_sankey(self,vector1, vector2):
        """
        Plot the sankey of the cells
        """
        return pl.plot_sankey(vector1, vector2)

    def plot_top_genes(self,**kwargs):
        """
        Plot the top genes of the cells
        """
        return pl.plot_top_genes(self.adata,**kwargs)
    # End-of-file (EOF)



import torch
import numpy as np

# Define performance metrics
def MSE(expr, expr_hat):
    return np.mean(np.square(expr_hat - expr))

def MAE(expr, expr_hat):
    return np.mean(np.abs(expr_hat - expr))

def RMSE(expr, expr_hat):
    return np.sqrt(np.mean(np.square(expr_hat - expr)))

def MAPE(expr, expr_hat, threshold=0.1):
    v = np.clip(np.abs(expr), threshold, None)
    diff = np.abs((expr_hat - expr) / v)
    return 100.0 * np.mean(diff)

def R2(expr, expr_hat):
    rss = np.sum(np.square(expr_hat - expr))
    tss = np.sum(np.square(expr - np.mean(expr)))
    return 1 - rss/tss

# Generate graph Laplacian for graph
def generate_graph_Laplacian(A, normalized = True):
    
    if np.allclose(A, np.eye(A.shape[0])):
        
        L = A
        
    else:
        if normalized:

            d = np.sum(A, axis=0)
            nz_index = np.where(d != 0)
            d[nz_index] = d[nz_index] ** (-0.5)
            d = np.expand_dims(d, axis=1)
            A = d.T * A * d

        L = np.diag(np.sum(A, axis=0)) - A
    
    L = torch.from_numpy(L.astype(np.float32))
    
    return L


#!/usr/bin/python -tt

"""
"""

import os,sys
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils
import torch.distributions
import random
import numpy as np
from sklearn import preprocessing
# from prettytable import PrettyTable

# from gaston.parse_args import args
import argparse
from gaston.neural_net import load_rescale_input_data,train
from gaston import dp_related


def main():
    # PARSE ARGS
    description="GASTON: interpretable deep learning model for mapping tissue topography"
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument('-o', '--output_layer', type=str, required=True, help="filename for N x G numpy array of GLM-PC values (column) for each spatial coordinate (row)")
    parser.add_argument('-i', '--input_layer', type=str, required=True, help="filename for N x 2 numpy array of position (x,y) of each spatial location in sample")
    parser.add_argument('-d', '--output_dir', type=str, required=False, default="./", help="The directory to save the output files")
    
    parser.add_argument('-e', '--epochs', type=int, required=False, default=10000, help="number of epochs to train the neural network")
    parser.add_argument('-c', '--checkpoint', type=int, required=False, default=500, help="save model every checkpoint epochs")

    # parser.add_argument('-u', '--hidden_units_spatial', type=int, required=False, default=50, help="number of hidden units encoding spatial coordinates to depth")
    # parser.add_argument('--hidden_layers_spatial', type=int, required=False, default=1, help="number of hidden layers encoding spatial coordinates to depth")
    parser.add_argument('-p', '--hidden_spatial', nargs='+', type=int, required=True, help="architecture of fully connected NN transforming spatial coordinates (x,y) to isodepth")

    # parser.add_argument('-x', '--hidden_units_expression', type=int, required=False, default=10, help="number of hidden units encoding depth to expression")
    # parser.add_argument('--hidden_layers_expression', type=int, required=False, default=1, help="number of hidden layers encoding depth to expression")
    parser.add_argument('-x', '--hidden_expression', nargs='+', type=int, required=True, help="architecture of fully connected NN transforming isodepth to expression GLM-PCs")

    parser.add_argument('-z', '--optimizer', type=str, required=False, default="adam", help="optimizer for fitting the neural network")
    parser.add_argument('-s', '--seed', type=int, required=False, default=0, help="Set random seed for reproducibility")

    args=parser.parse_args(sys.argv[1:])

    ###### RUN NN
    out_dir_seed=f"{args.output_dir}/seed{args.seed}" # save in rep{seed}
    os.makedirs(out_dir_seed, exist_ok=True) 
    # Load input data and rescale, S_torch and A_torch are torch tensors representing the spatial coordinates and expression data, respectively
    S=np.load(args.input_layer)
    A=np.load(args.output_layer)
    S_torch, A_torch = load_rescale_input_data(S,A)

    # train neural net
    mod, loss_list = train(S_torch, A_torch,
                          S_hidden_list=args.hidden_spatial, A_hidden_list=args.hidden_expression, 
                          epochs=args.epochs, checkpoint=args.checkpoint, 
                          save_dir=out_dir_seed, optim=args.optimizer, seed=args.seed, save_final=True) 

def get_parser():
    description = "GASTON"
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument('-o', '--output_layer', type=str, required=True, help="filename for N x G numpy array of GLM-PC values (column) for each spatial coordinate (row)")
    parser.add_argument('-i', '--input_layer', type=str, required=True, help="filename for N x 2 numpy array of position (x,y) of each spatial location in sample")
    parser.add_argument('-d', '--output_dir', type=str, required=False, default="./", help="The directory to save the output files")
    
    parser.add_argument('-e', '--epochs', type=int, required=False, default=10000, help="number of epochs to train the neural network")
    parser.add_argument('-c', '--checkpoint', type=int, required=False, default=500, help="save model every checkpoint epochs")

    # parser.add_argument('-u', '--hidden_units_spatial', type=int, required=False, default=50, help="number of hidden units encoding spatial coordinates to depth")
    # parser.add_argument('--hidden_layers_spatial', type=int, required=False, default=1, help="number of hidden layers encoding spatial coordinates to depth")
    parser.add_argument('-p', '--hidden_spatial', nargs='+', type=int, required=True, help="architecture of fully connected NN transforming spatial coordinates (x,y) to isodepth")

    # parser.add_argument('-x', '--hidden_units_expression', type=int, required=False, default=10, help="number of hidden units encoding depth to expression")
    # parser.add_argument('--hidden_layers_expression', type=int, required=False, default=1, help="number of hidden layers encoding depth to expression")
    parser.add_argument('-x', '--hidden_expression', nargs='+', type=int, required=True, help="architecture of fully connected NN transforming isodepth to expression GLM-PCs")

    parser.add_argument('-z', '--optimizer', type=str, required=False, default="adam", help="optimizer for fitting the neural network")
    parser.add_argument('-s', '--seed', type=int, required=False, default=0, help="Set random seed for reproducibility")

    return parser


if __name__ == '__main__':
    main()




# def count_parameters(model):
#     table = PrettyTable(["Modules", "Parameters"])
#     total_params = 0
#     for name, parameter in model.named_parameters():
#         if not parameter.requires_grad: continue
#             param = parameter.numel()
#             table.add_row([name, param])
#             total_params+=param
#   print(table)
#   print(f"Total Trainable Params: {total_params}")
#   return total_params


import numpy as np


def log_transform(ad, ps=0.1):
    """Compute log transformation of AnnData data in place.

    :param ad: anndata.AnnData object to be normalized
    :param ps: (float) pseudo-count for log transformation to avoid log(0) errors
    """
    ad.X = np.log2(ad.X + ps) - np.log2(ps)





from ..externel.tosica import fit_model
from ..externel.tosica.TOSICA_model import scTrans_model as create_model
from ..externel.tosica.train import set_seed,splitDataSet,get_gmt,read_gmt,create_pathway_mask,train_one_epoch,evaluate,MyDataSet
from ..externel.tosica.pre import predict,predicted,todense
import random
import torch.nn.functional as F
import numpy as np

from torch.utils.data import Dataset

import sys
import pandas as pd
from tqdm import tqdm
from sklearn.preprocessing import LabelEncoder
from collections import OrderedDict
import scanpy as sc
import anndata as ad

import os
import math
import torch
import torch.optim as optim
import torch.optim.lr_scheduler as lr_scheduler
from torch.utils.tensorboard import SummaryWriter
import time
import anndata
from .._settings import add_reference



class pyTOSICA(object):

    def __init__(self,adata:anndata.AnnData,project_path:str,gmt_path=None,
                 label_name:str='Celltype',mask_ratio:float=0.015,
                 max_g:int=300,max_gs:int=300,n_unannotated:int= 1,
                 embed_dim:int=48,depth:int=1,num_heads:int=4,batch_size:int=8,
                 device:str='cuda:0'
                 ) -> None:
        """
        Initialize a pyTOSICA object.

        Arguments:
            adata: AnnData object.
            project_path: Path to save the results.
            gmt_path: The name (human_gobp; human_immune; human_reactome; human_tf; mouse_gobp; mouse_reactome and mouse_tf) or path of mask to be used.
            label_name: The column name of the label you want to prediect. Should in adata.obs.columns.
            mask_ratio: The ratio of the connection reserved when there is no available mask.
            max_g: The maximum number of genes to be used.
            max_gs: The maximum number of pathways to be used.
            n_unannotated: The number of unannotated genes to be added.
            embed_dim: The dimension of the embedding.
            depth: The depth of the model.
            num_heads: The number of heads in the model.
            batch_size: The batch size.
            device: The device to be used.

        
        """


        self.adata=adata 
        self.gmt_path=gmt_path
        self.project_path=project_path
        self.label_name=label_name
        self.n_unannotated=n_unannotated
        self.embed_dim=embed_dim

        GLOBAL_SEED = 1
        set_seed(GLOBAL_SEED)
        #device = 'cuda:0'
        device = torch.device(device if torch.cuda.is_available() else "cpu")
        self.device = device
        print(device)
        today = time.strftime('%Y%m%d',time.localtime(time.time()))
        #train_weights = os.getcwd()+"/weights%s"%today
        #project = project or gmt_path.replace('.gmt','')+'_%s'%today
        #project_path = os.getcwd()+'/%s'%project
        if os.path.exists(project_path) is False:
            os.makedirs(project_path)
        self.tb_writer = SummaryWriter()
        exp_train, label_train, exp_valid, label_valid, inverse,genes = splitDataSet(adata,label_name)
        if gmt_path is None:
            mask = np.random.binomial(1,mask_ratio,size=(len(genes), max_gs))
            pathway = list()
            for i in range(max_gs):
                x = 'node %d' % i
                pathway.append(x)
            print('Full connection!')
        else:
            if '.gmt' in gmt_path:
                gmt_path = gmt_path
            else:
                gmt_path = get_gmt(gmt_path)
                if gmt_path=="Error":
                    print("You need to download the gene sets first using ov.utils.download_tosica_gmt()")
            
            reactome_dict = read_gmt(gmt_path, min_g=0, max_g=max_g)
            mask,pathway = create_pathway_mask(feature_list=genes,
                                            dict_pathway=reactome_dict,
                                            add_missing=n_unannotated,
                                            fully_connected=True)
            pathway = pathway[np.sum(mask,axis=0)>4]
            mask = mask[:,np.sum(mask,axis=0)>4]
            #print(mask.shape)
            pathway = pathway[sorted(np.argsort(np.sum(mask,axis=0))[-min(max_gs,mask.shape[1]):])]
            mask = mask[:,sorted(np.argsort(np.sum(mask,axis=0))[-min(max_gs,mask.shape[1]):])]
            #print(mask.shape)
            print('Mask loaded!')
        np.save(project_path+'/mask.npy',mask)
        pd.DataFrame(pathway).to_csv(project_path+'/pathway.csv') 
        pd.DataFrame(inverse,columns=[label_name]).to_csv(project_path+'/label_dictionary.csv', quoting=None)
        self.pathway=pathway 
        self.mask=mask

        num_classes = np.int64(torch.max(label_train)+1)

        train_dataset = MyDataSet(exp_train, label_train)
        valid_dataset = MyDataSet(exp_valid, label_valid)
        self.train_loader = torch.utils.data.DataLoader(train_dataset,
                                                batch_size=batch_size,
                                                shuffle=True,
                                                pin_memory=True,drop_last=True)
        self.valid_loader = torch.utils.data.DataLoader(valid_dataset,
                                                batch_size=batch_size,
                                                shuffle=False,
                                                pin_memory=True,drop_last=True)

        self.model = create_model(num_classes=num_classes, 
                                  num_genes=len(exp_train[0]),  
                                  mask = mask,embed_dim=embed_dim,
                                  depth=depth,num_heads=num_heads,
                                  has_logits=False).to(device) 
        pass

    def train(self,pre_weights:str='',lr:float=0.001, epochs:int= 10, lrf:float=0.01):
        """
        Trainning the tosica model.

        Arguments:
            pre_weights: The path of the pre-trained weights.
            lr: The learning rate.
            epochs: The number of epochs.
            lrf: The learning rate of the last layer.

        Returns:
            model: The trained model.
        
        """
        if pre_weights != "":
            assert os.path.exists(pre_weights), "pre_weights file: '{}' not exist.".format(pre_weights)
            preweights_dict = torch.load(pre_weights, map_location=self.device)
            print(self.model.load_state_dict(preweights_dict, strict=False))
        #for name, param in model.named_parameters():
        #    if param.requires_grad:
        #        print(name) 
        print('Model builded!')
        pg = [p for p in self.model.parameters() if p.requires_grad]  
        optimizer = optim.SGD(pg, lr=lr, momentum=0.9, weight_decay=5E-5) 
        lf = lambda x: ((1 + math.cos(x * math.pi / epochs)) / 2) * (1 - lrf) + lrf  
        scheduler = lr_scheduler.LambdaLR(optimizer, lr_lambda=lf)
        for epoch in range(epochs):
            train_loss, train_acc = train_one_epoch(model=self.model,
                                                    optimizer=optimizer,
                                                    data_loader=self.train_loader,
                                                    device=self.device,
                                                    epoch=epoch)
            scheduler.step() 
            val_loss, val_acc = evaluate(model=self.model,
                                        data_loader=self.valid_loader,
                                        device=self.device,
                                        epoch=epoch)
            tags = ["train_loss", "train_acc", "val_loss", "val_acc", "learning_rate"]
            self.tb_writer.add_scalar(tags[0], train_loss, epoch)
            self.tb_writer.add_scalar(tags[1], train_acc, epoch)
            self.tb_writer.add_scalar(tags[2], val_loss, epoch)
            self.tb_writer.add_scalar(tags[3], val_acc, epoch)
            self.tb_writer.add_scalar(tags[4], optimizer.param_groups[0]["lr"], epoch)
            torch.save(self.model.state_dict(), self.project_path+"/model-{}.pth".format(epoch))
        print('Training finished!')
        add_reference(self.adata,'TOSICA','cell type classification with TOSICA')
        return self.model
    
    def save(self,save_path=None):
        """
        Save the model.

        Arguments:
            save_path: The path of the saved model.
        
        """
        if save_path==None:
            save_path = self.project_path+'/model-best.pth'
        torch.save(self.model.state_dict(), save_path)
        print('Model saved!')

    def load(self,load_path=None):
        """
        Load the model.

        Arguments:
            load_path: The path of the loaded model.

        """
        if load_path==None:
            load_path = self.project_path+'/model-best.pth'
        self.model.load_state_dict(torch.load(load_path, map_location=self.device))
        print('Model loaded!')
        
        
    def predicted(self,pre_adata:anndata.AnnData,laten:bool=False,n_step:int=10000,cutoff:float=0.1,
        batch_size:int=50,):
        """
        Predict the cell type of the new data.

        Arguments:
            pre_adata: The new data.
            laten: Whether to get the latent representation.
            n_step: The number of steps of the random walk.
            cutoff: The cutoff of the random walk.
            batch_size: The batch size of the random walk.

        Returns:
            new_adata: The new data with predicted cell type.
        
        """

        mask_path = os.getcwd()+'/%s'%self.project_path+'/mask.npy'
        dictionary = pd.read_table(self.project_path+'/label_dictionary.csv', sep=',',header=0,index_col=0)
        self.pathway=pd.read_csv(self.project_path+'/pathway.csv', index_col=0)
        n_c = len(dictionary)
        dic = {}
        for i in range(len(dictionary)):
            dic[i] = dictionary[self.label_name][i]
        self.model.eval()
        parm={}
        for name,parameters in self.model.named_parameters():
            #print(name,':',parameters.size())
            parm[name]=parameters.detach().cpu().numpy()
        gene2token = parm['feature_embed.fe.weight']
        gene2token = gene2token.reshape((int(gene2token.shape[0]/self.embed_dim),self.embed_dim,pre_adata.shape[1]))
        gene2token = abs(gene2token)
        gene2token = np.max(gene2token,axis=1)
        gene2token = pd.DataFrame(gene2token)
        gene2token.columns=pre_adata.var_names
        gene2token.index = self.pathway['0']
        gene2token.to_csv(self.project_path+'/gene2token_weights.csv')
        latent = torch.empty([0,self.embed_dim]).cpu()
        att = torch.empty([0,(len(self.pathway))]).cpu()
        predict_class = np.empty(shape=0)
        pre_class = np.empty(shape=0)      
        latent = torch.squeeze(latent).cpu().numpy()
        l_p = np.c_[latent, predict_class,pre_class]
        att = np.c_[att, predict_class,pre_class]
        all_line = pre_adata.shape[0]
        n_line = 0
        adata_list = []
        while (n_line) <= all_line:
            if (all_line-n_line)%batch_size != 1:
                expdata = pd.DataFrame(todense(pre_adata[n_line:n_line+min(n_step,(all_line-n_line))]),
                                       index=np.array(pre_adata[n_line:n_line+min(n_step,(all_line-n_line))].obs_names).tolist(), 
                                       columns=np.array(pre_adata.var_names).tolist())
                print(n_line)
                n_line = n_line+n_step
            else:
                expdata = pd.DataFrame(todense(pre_adata[n_line:n_line+min(n_step,(all_line-n_line-2))]),
                                       index=np.array(pre_adata[n_line:n_line+min(n_step,(all_line-n_line-2))].obs_names).tolist(), 
                                       columns=np.array(pre_adata.var_names).tolist())
                n_line = (all_line-n_line-2)
                print(n_line)
            expdata = np.array(expdata)
            expdata = torch.from_numpy(expdata.astype(np.float32))
            data_loader = torch.utils.data.DataLoader(expdata,
                                                    batch_size=batch_size,
                                                    shuffle=False,
                                                    pin_memory=True)
            with torch.no_grad():
                # predict class
                for step, data in enumerate(data_loader):
                    #print(step)
                    exp = data
                    lat, pre, weights = self.model(exp.to(self.device))
                    pre = torch.squeeze(pre).cpu()
                    pre = F.softmax(pre,1)
                    predict_class = np.empty(shape=0)
                    pre_class = np.empty(shape=0) 
                    for i in range(len(pre)):
                        if torch.max(pre, dim=1)[0][i] >= cutoff: 
                            predict_class = np.r_[predict_class,torch.max(pre, dim=1)[1][i].numpy()]
                        else:
                            predict_class = np.r_[predict_class,n_c]
                        pre_class = np.r_[pre_class,torch.max(pre, dim=1)[0][i]]     
                    l_p = torch.squeeze(lat).cpu().numpy()
                    att = torch.squeeze(weights).cpu().numpy()
                    meta = np.c_[predict_class,pre_class]
                    meta = pd.DataFrame(meta)
                    meta.columns = ['Prediction','Probability']
                    meta.index = meta.index.astype('str')
                    if laten:
                        l_p = l_p.astype('float32')
                        new = sc.AnnData(l_p, obs=meta)
                    else:
                        att = att[:,0:(len(self.pathway)-self.n_unannotated)]
                        att = att.astype('float32')
                        varinfo = pd.DataFrame(self.pathway.iloc[0:len(self.pathway)-self.n_unannotated,0].values,
                                               index=self.pathway.iloc[0:len(self.pathway)-self.n_unannotated,0],columns=['pathway_index'])
                        new = sc.AnnData(att, obs=meta, var = varinfo)
                    adata_list.append(new)
        print(all_line)
        new = ad.concat(adata_list)
        new.obs.index = pre_adata.obs.index
        new.obs['Prediction'] = new.obs['Prediction'].map(dic)
        new.obs[pre_adata.obs.columns] = pre_adata.obs[pre_adata.obs.columns].values    
        add_reference(self.adata,'TOSICA','cell type classification with TOSICA')

        return(new)


def train(adata, gmt_path, project=None,pre_weights='', 
          label_name='Celltype',max_g=300,max_gs=300,
          mask_ratio =0.015, n_unannotated = 1,
          batch_size=8, embed_dim=48,depth=2,
          num_heads=4,lr=0.001, epochs= 10, lrf=0.01):
    r"""
    Fit the model with reference data
    
    Arguments:
        adatas: Single-cell datasets
        gmt_path: The name (human_gobp; human_immune; human_reactome; human_tf; mouse_gobp; mouse_reactome and mouse_tf) or path of mask to be used.
        project: The name of project. Default: gmt_path_today.
        pre_weights: The path to the pre-trained weights. If pre_weights = '', the model will be trained from scratch.
        label_name: The column name of the label you want to prediect. Should in adata.obs.columns.
        max_g: The max of gene number belong to one pathway.
        max_gs: The max of pathway/token number.
        mask_ratio: The ratio of the connection reserved when there is no available mask.
        n_unannotated: The number of fully connected tokens to be added.
        batch_size: The number of cells for training in one epoch.
        embed_dim: The dimension of pathway/token embedding.
        depth: The number of multi-head self-attention layer.
        num_heads: The number of head in one self-attention layer.
        lr: Learning rate.
        epochs: The number of epoch will be trained.
        lrf: The hyper-parameter of Cosine Annealing.
    
    Returns:
        ./mask.npy: Mask matrix
        ./pathway.csv: Gene set list
        ./label_dictionary.csv: Label list
        ./weights20220603/: Weights
    
    """

    fit_model(adata, gmt_path, project=project,pre_weights=pre_weights, label_name=label_name,
              max_g=max_g,max_gs=max_gs,mask_ratio=mask_ratio, n_unannotated = n_unannotated,batch_size=batch_size, 
              embed_dim=embed_dim,depth=depth,num_heads=num_heads,lr=lr, epochs= epochs, lrf=lrf)


def pre(adata,model_weight_path,project,laten=False,save_att = 'X_att', 
        save_lantent = 'X_lat',n_step=10000,cutoff=0.1,
        n_unannotated = 1,batch_size=50,embed_dim=48,depth=2,num_heads=4):
    r"""
    Prediect query data with the model and pre-trained weights.
    
    Arguments:
        adatas: Query single-cell datasets.
        model_weight_path: The path to the pre-trained weights.
        mask_path: The path to the mask matrix.
        project: The name of project.
        laten: Get laten output.
        save_att: The name of the attention matrix to be added in the adata.obsm.
        save_lantent: The name of the laten matrix to be added in the adata.obsm.
        max_gs: The max of pathway/token number.
        n_step: The number of cells load into memory at the same time.
        cutoff: Unknown cutoff.
        n_unannotated: The number of fully connected tokens to be added. Should be the same as train.
        batch_size: The number of cells for training in one epoch.
        embed_dim: The dimension of pathway/token embedding. Should be the same as train.
        depth: The number of multi-head self-attention layer. Should be the same as train.
        num_heads: The number of head in one self-attention layer. Should be the same as train.
    
    Returns:
        adata: adata.X : Attention matrix
        adata.obs['Prediction'] : Predicted labels
        adata.obs['Probability'] : Probability of the prediction
        adata.var['pathway_index'] : Gene set of each colume
    """
    
    mask_path = os.getcwd()+'/%s'%project+'/mask.npy'
    adata = predict(adata,model_weight_path,project=project,mask_path = mask_path,laten=laten,
             save_att = save_att, save_lantent = save_lantent,n_step=n_step,cutoff=cutoff,n_unannotated = n_unannotated,batch_size=batch_size,embed_dim=embed_dim,depth=depth,num_heads=num_heads)
    return(adata)


import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
from torch.autograd import Variable
from tqdm import tqdm
from torch.optim import AdamW
from torch.utils.data import Dataset, DataLoader
import pandas as pd
import numpy as np
from math import exp
import copy


class VAE(nn.Module):
    def __init__(self, embedding_size, hidden_size_list: list, mid_hidden):
        super(VAE, self).__init__()
        self.embedding_size = embedding_size
        self.hidden_size_list = hidden_size_list
        self.mid_hidden = mid_hidden

        self.enc_feature_size_list = [self.embedding_size] + self.hidden_size_list + [self.mid_hidden * 2]
        self.dec_feature_size_list = [self.embedding_size] + self.hidden_size_list + [self.mid_hidden]

        self.encoder = nn.ModuleList(
            [nn.Linear(self.enc_feature_size_list[i], self.enc_feature_size_list[i + 1]) for i in
             range(len(self.enc_feature_size_list) - 1)])
        self.decoder = nn.ModuleList(
            [nn.Linear(self.dec_feature_size_list[i], self.dec_feature_size_list[i - 1]) for i in
             range(len(self.dec_feature_size_list) - 1, 0, -1)])

    def encode(self, x):
        for i, layer in enumerate(self.encoder):
            x = self.encoder[i](x)
            if i != len(self.encoder) - 1:
                x = F.relu(x)
        return x

    def decode(self, x):
        for i, layer in enumerate(self.decoder):
            x = self.decoder[i](x)
            x = F.relu(x)
        return x

    def forward(self, x, used_device):
        x = x.to(used_device)
        encoder_output = self.encode(x)
        mu, sigma = torch.chunk(encoder_output, 2, dim=1)  # mu, log_var
        hidden = torch.randn_like(sigma) + mu * torch.exp(sigma) ** 0.5  # var => std
        x_hat = self.decode(hidden)
        kl_div = 0.5 * torch.sum(torch.exp(sigma) + torch.pow(mu, 2) - 1 - sigma) / (x.shape[0] * x.shape[1])
        return x_hat, kl_div

    def get_hidden(self, x):
        encoder_output = self.encode(x)
        mu, sigma = torch.chunk(encoder_output, 2, dim=1)  # mu, log_var
        hidden = torch.randn_like(sigma) * torch.exp(sigma) ** 0.5 + mu  # var => std
        return hidden


# bulk deconvolution
class BulkDataset(Dataset):  
    def __init__(self, single_cell, label):
        self.sc = single_cell
        self.label = label

    def __getitem__(self, idx):
        tmp_x = self.sc[idx]
        tmp_y_tag = self.label[idx]

        return (tmp_x, tmp_y_tag) 

    def __len__(self):
        return self.label.shape[0]


def train_vae(single_cell, label, used_device, batch_size, feature_size, epoch_num, learning_rate, hidden_size, patience=10):
    batch_size = batch_size
    feature_size = feature_size
    epoch_num = epoch_num
    lr = learning_rate
    hidden_list = [2048, 1024, 512]

    mid_hidden_size = hidden_size
    weight_decay = 5e-4
    dataloader = DataLoader(BulkDataset(single_cell=single_cell, label=label), batch_size=batch_size, shuffle=True,
                            pin_memory=True)
    criterion = nn.MSELoss()
    beta = 4
    beta1 = 0.9
    beta2 = 0.999
    vae = VAE(feature_size, hidden_list, mid_hidden_size).to(used_device)
    optimizer = AdamW(vae.parameters(), lr=lr, weight_decay=weight_decay, betas=(beta1, beta2))

    pbar = tqdm(range(epoch_num))
    min_loss = 1000000000000000
    vae.train()
    early_stop = 0
    history = []

    for epoch in pbar:
        train_loss = 0

        for batch_idx, data in enumerate(dataloader):
            cell_feature, label = data
            cell_feature = torch.tensor(cell_feature, dtype=torch.float32).to(used_device)
            x_recon, total_kld = vae(cell_feature, used_device)

            recon_loss = criterion(x_recon, cell_feature)
            beta_vae_loss = recon_loss + beta * total_kld
            loss = beta_vae_loss

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            train_loss += loss.item()

        if train_loss < min_loss:
            min_loss = train_loss
            best_vae = copy.deepcopy(vae)
            early_stop = 0
            epoch_final = epoch  
        else:
            early_stop += 1

        history.append(train_loss)
        if early_stop == patience:
            print('Early stopping at epoch {}...'.format(epoch+1))
            break
        pbar.set_description('Train Epoch: {}'.format(epoch))
        pbar.set_postfix(loss=f"{train_loss:.4f}", min_loss=f"{min_loss:.4f}")
    print(f"min loss = {min_loss}")
    
    return best_vae,history


def load_vae(feature_size, hidden_size, path, used_device):
    hidden_list = [2048, 1024, 512]
    vae = VAE(feature_size, hidden_list, hidden_size).to(used_device)
    vae.load_state_dict(torch.load(path, map_location=used_device))
    
    return vae


def generate_vae(net, ratio, single_cell, label, breed_2_list, index_2_gene, cell_number_target_num=None,
                 used_device=None):
    # net in cuda now
    for p in net.parameters():  # reset requires_grad
        p.requires_grad = False  # avoid computation

    net.eval()
    net.to(used_device)
    cell_all_generate = []
    label_all_generate = []

    all_to_generate = 0
    for x in cell_number_target_num.values():
        all_to_generate += x

    if cell_number_target_num != None:
        epochs = 10000 
        ratio = 1
    else:
        epochs = 1

    cell_feature = torch.from_numpy(single_cell).float()
    label = torch.from_numpy(label)

    with torch.no_grad():
        with tqdm(total=all_to_generate, desc='generating') as pbar:
            for epoch in range(epochs):
                key_list = []  # list
                generate_num = 0

                label_list = label.tolist()
                for i in range(len(label_list)):
                    if cell_number_target_num[label_list[i]] <= 0:
                        continue
                    else:
                        cell_number_target_num[label_list[i]] -= 1
                        generate_num += 1
                        key_list.append(i)

                if cell_number_target_num == None or all_to_generate == 0 or len(key_list) == 0:
                    assert all_to_generate == 0 and len(key_list) == 0
                    break

                import random
                random.shuffle(key_list)

                label = label.index_select(0, torch.tensor(key_list))
                cell_feature = cell_feature.index_select(0, torch.tensor(key_list))

                dataloader = DataLoader(BulkDataset(single_cell=cell_feature, label=label), batch_size=300,
                                        shuffle=False,
                                        pin_memory=True, num_workers=0)
                for batch_idx, data in enumerate(dataloader):  
                    cell_feature_batch, label_batch = data
                    cell_feature_batch = cell_feature_batch.to(used_device)

                    label_batch = label_batch.cpu().numpy()

                    for j in range(ratio): 
                        ans_l, _ = net(cell_feature_batch, used_device)
                        ans_l = ans_l.cpu().data.numpy()
                        cell_all_generate.extend(ans_l)
                        label_all_generate.extend(label_batch)

                all_to_generate -= generate_num
                pbar.update(generate_num)

    print("generated done!")
    generate_sc_meta, generate_sc_data = prepare_data(cell_all_generate, label_all_generate, breed_2_list,
                                                      index_2_gene)
    return generate_sc_meta, generate_sc_data


def prepare_data(cell_all_generate, label_all_generate, breed_2_list, index_2_gene):
    cell_all_generate = np.array(cell_all_generate)
    label_all_generate = np.array(label_all_generate)

    cell_all_generate_csv = pd.DataFrame(cell_all_generate)
    label_all_generate_csv = pd.DataFrame(label_all_generate)

    ids = label_all_generate_csv[0].tolist()
    breeds = []
    for id in ids:
        breeds.append(breed_2_list[id])
    name = ["C_" + str(i + 1) for i in range(label_all_generate.shape[0])]

    label_all_generate_csv.insert(1, "Cell_type", np.array(breeds))
    label_all_generate_csv.insert(1, "Cell", np.array(name))
    label_all_generate_csv = label_all_generate_csv.drop([0], axis=1)

    cell_all_generate_csv = cell_all_generate_csv.T
    cell_all_generate_csv.columns = name
    cell_all_generate_csv.index = index_2_gene

    return label_all_generate_csv, cell_all_generate_csv



import argparse
import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.neighbors import NearestNeighbors
import os

def load_data(filepath):
    if not os.path.exists(filepath):
        raise FileNotFoundError("File not found.")
    
    # Load data
    data = pd.read_csv(filepath, sep="\t", header=0)
    
    # Set row names and clean them
    data.index = data.iloc[:, 0]
    data.index = data.index.str.replace('.', '-', regex=False)
    data = data.iloc[:, 1:]  # Remove the first column
    
    return data

def smooth_data_kNN(output_dir, suffix, max_pcs, seed):
    np.random.seed(seed)

    # Load data
    predicted_df = load_data(os.path.join(output_dir, f'binned_df{suffix}.txt'))
    ranked_df = load_data(os.path.join(output_dir, f'ranked_df{suffix}.txt'))
    with open(os.path.join(output_dir, f'top_var_genes{suffix}.txt'), 'r') as f:
        top_genes = f.read().splitlines()

    def smoothing_by_KNN(score, umap_coordinates):
        maxk = min(30, max(3, round(0.005 * len(score))))
        nbrs = NearestNeighbors(n_neighbors=maxk).fit(umap_coordinates)
        _, indices = nbrs.kneighbors(umap_coordinates)
        smoothed_scores = [np.median(score[indices[i]]) for i in range(len(score))]
        return smoothed_scores

    # Convert ranked_df to numpy array
    ranked_data = ranked_df.to_numpy()

    # Perform PCA
    pca = PCA(n_components=min(ranked_data.shape[1] - 1, max_pcs))
    pca_result = pca.fit_transform(ranked_data)
    
    # Calculate explained variance
    explained_variance_ratio = pca.explained_variance_ratio_
    cumulative_variance = np.cumsum(explained_variance_ratio)
    threshold_var_explained = 0.5
    num_pcs = np.argmax(cumulative_variance >= threshold_var_explained) + 1
    num_pcs = max(2, num_pcs)

    umap_coordinates = pca_result[:, :num_pcs]

    # Apply kNN smoothing
    knn_score = smoothing_by_KNN(predicted_df['preKNN_CytoTRACE2_Score'], umap_coordinates)
    predicted_df['CytoTRACE2_Score'] = knn_score

    # Re-map potency categories
    ranges = np.linspace(0, 1, 7)
    labels = [
        'Differentiated',
        'Unipotent',
        'Oligopotent',
        'Multipotent',
        'Pluripotent',
        'Totipotent'
    ]
    order_vector = pd.cut(predicted_df['CytoTRACE2_Score'], bins=ranges, labels=labels, include_lowest=True)
    predicted_df['CytoTRACE2_Potency'] = order_vector

    # Save the result
    predicted_df.to_csv(os.path.join(output_dir, f'smoothbykNNresult{suffix}.txt'), sep='\t', index=False, quoting=3)

# Argument parsing
parser = argparse.ArgumentParser()

parser.add_argument("--output-dir", type=str, default="cytotrace2_results", help="Output directory containing intermediate files")
parser.add_argument("--suffix", type=str, default="", help="Suffix of intermediate files")
parser.add_argument("--max-pcs", type=int, default=200, help="Integer, indicating the maximum number of principal components to use in the smoothing by kNN step (default is 200)")
parser.add_argument("--seed", type=int, default=14, help="Integer, specifying the seed for reproducibility in random processes (default is 14).")

args = parser.parse_args()
smooth_data_kNN(output_dir=args.output_dir, suffix=args.suffix, max_pcs=args.max_pcs, seed=args.seed)



import numpy as np
import scipy.sparse as sp
import torch
import torch.nn as nn
import torch.nn.functional as F

from ..nn.gcn import sparse_or_dense_dropout
from ..utils import to_sparse_tensor

__all__ = [
    'ImprovedGCN',
    'ImpGraphConvolution',
]


class ImpGraphConvolution(nn.Module):
    """Graph convolution layer.

    Args:
        in_features: Size of each input sample.
        out_features: Size of each output sample.

    """
    def __init__(self, in_features, out_features):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight_own = nn.Parameter(torch.empty(in_features, out_features))
        self.weight_nbr = nn.Parameter(torch.empty(in_features, out_features))
        self.bias = nn.Parameter(torch.empty(out_features))
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.weight_own, gain=2.0)
        nn.init.xavier_uniform_(self.weight_nbr, gain=2.0)
        nn.init.zeros_(self.bias)

    def forward(self, x, adj):
        return adj @ (x @ self.weight_nbr) + x @ self.weight_own + self.bias


class ImprovedGCN(nn.Module):
    """An improved GCN architecture.

    This version uses two weight matrices for self-propagation and aggregation,
    doesn't use batchnorm, and uses Tanh instead of ReLU nonlinearities.
    Has more stable training / faster convergence than standard GCN for overlapping
    community detection.

    This improved architecture was inspired by https://arxiv.org/abs/1906.12192
    """
    def __init__(self, input_dim, hidden_dims, output_dim, dropout=0.5, layer_norm=False):
        super().__init__()
        self.dropout = dropout
        layer_dims = np.concatenate([hidden_dims, [output_dim]]).astype(np.int32)
        self.layers = nn.ModuleList([ImpGraphConvolution(input_dim, layer_dims[0])])
        for idx in range(len(layer_dims) - 1):
            self.layers.append(ImpGraphConvolution(layer_dims[idx], layer_dims[idx + 1]))
        if layer_norm:
            self.layer_norm = [
                nn.LayerNorm([dim], elementwise_affine=False) for dim in hidden_dims
            ]
        else:
            self.layer_norm = None

    @staticmethod
    def normalize_adj(adj : sp.csr_matrix):
        """Normalize adjacency matrix and convert it to a sparse tensor."""
        if sp.isspmatrix(adj):
            adj = adj.tolil()
            adj.setdiag(0)
            adj = adj.tocsr()
            deg = np.ravel(adj.sum(1))
            deg_sqrt_inv = 1 / np.sqrt(deg)
            adj_norm = adj.multiply(deg_sqrt_inv[:, None]).multiply(deg_sqrt_inv[None, :])
        elif torch.is_tensor(adj):
            deg = adj.sum(1)
            deg_sqrt_inv = 1 / torch.sqrt(deg)
            adj_norm = adj * deg_sqrt_inv[:, None] * deg_sqrt_inv[None, :]
        return to_sparse_tensor(adj_norm)

    def forward(self, x, adj):
        for idx, gcn in enumerate(self.layers):
            if self.dropout != 0:
                x = sparse_or_dense_dropout(x, p=self.dropout, training=self.training)
            x = gcn(x, adj)
            if idx != len(self.layers) - 1:
                x = torch.tanh(x)
                if self.layer_norm is not None:
                    x = self.layer_norm[idx](x)
        return x

    def get_weights(self):
        """Return the weight matrices of the model."""
        return [w for n, w in self.named_parameters() if 'bias' not in n]

    def get_biases(self):
        """Return the bias vectors of the model."""
        return [w for n, w in self.named_parameters() if 'bias' in n]



# OmicVerse Analytics Configuration
"""
OmicVerse 统计追踪配置

这个模块提供了对报告使用情况的统计追踪功能。
所有数据都是匿名的，不会收集个人隐私信息。

收集的信息包括：
- 报告生成时间
- 匿名用户ID（基于机器信息的hash）
- 操作系统信息
- 浏览器类型（仅在打开报告时）
- 地理位置（仅国家/地区级别）
"""

# 默认配置
DEFAULT_CONFIG = {
    # 是否启用统计
    'enable_analytics': True,
    
    # 统计服务器地址
    'analytics_endpoint': 'https://analytics.omicverse.org/track.gif',
    
    # 备用Google Analytics ID（如果需要）
    'google_analytics_id': None,
    
    # 项目标识符
    'project_id': 'omicverse-scrna',
    
    # 数据保留策略
    'data_retention_days': 365,
    
    # 隐私设置
    'privacy_mode': True,  # 启用隐私保护模式
    'hash_user_info': True,  # 对用户信息进行hash处理
    'collect_ip': False,  # 不收集IP地址
    'collect_detailed_ua': False,  # 不收集详细的用户代理信息
}

# 统计事件类型
EVENT_TYPES = {
    'REPORT_GENERATED': 'report_generated',
    'REPORT_VIEWED': 'report_viewed',
    'SECTION_VIEWED': 'section_viewed',
    'THEME_SWITCHED': 'theme_switched',
}

# 统计字段定义
TRACKED_FIELDS = {
    'timestamp': 'ISO格式时间戳',
    'event_type': '事件类型',
    'user_hash': '匿名用户标识',
    'session_id': '会话ID',
    'report_id': '报告标识',
    'platform': '操作系统平台',
    'language': '浏览器语言',
    'timezone': '时区',
    'screen_resolution': '屏幕分辨率',
    'viewport_size': '浏览器窗口大小',
}

def get_privacy_notice():
    """获取隐私声明"""
    return """
    📊 统计说明：
    
    OmicVerse 会收集匿名的使用统计信息，用于：
    • 了解软件使用情况
    • 改进软件功能
    • 提供更好的用户体验
    
    我们承诺：
    ✅ 所有数据都是匿名的
    ✅ 不收集个人隐私信息
    ✅ 不收集具体文件内容
    ✅ 遵循 GDPR 等隐私法规
    
    您可以通过设置 enable_analytics=False 来禁用统计功能。
    """

def get_opt_out_instructions():
    """获取退出统计的说明"""
    return """
    如何禁用统计：
    
    在生成报告时设置：
    ```python
    ov.generate_scRNA_report(
        adata, 
        enable_analytics=False
    )
    ```
    
    或设置环境变量：
    ```bash
    export OMICVERSE_ANALYTICS=false
    ```
    """ 